import sys, time, os

from confapp import conf


if not hasattr(conf,'PYFORMS_MODE'):
    from kivy.logger import Logger as logger
else:
    import logging; logger = logging.getLogger(__name__)

import copy
import numpy as np
import tensorflow as tf

from idtrackerai.postprocessing.identify_non_assigned_with_interpolation import assign_zeros_with_interpolation_identities
from idtrackerai.network.identification_model.store_accuracy_and_loss    import Store_Accuracy_and_Loss
from idtrackerai.postprocessing.correct_impossible_velocity_jumps        import correct_impossible_velocity_jumps
from idtrackerai.network.identification_model.network_params             import NetworkParams
from idtrackerai.postprocessing.compute_velocity_model                   import compute_model_velocity
from idtrackerai.network.identification_model.id_CNN                     import ConvNetwork
from idtrackerai.postprocessing.get_trajectories                         import produce_output_dict
from idtrackerai.postprocessing.assign_them_all                          import close_trajectories_gaps
from idtrackerai.accumulation_manager                                    import AccumulationManager
from idtrackerai.list_of_blobs                                           import ListOfBlobs
from idtrackerai.accumulator                                             import perform_one_accumulation_step
from idtrackerai.pre_trainer                                             import pre_train_global_fragment
from idtrackerai.assigner                                                import assign_remaining_fragments
from confapp                                                             import conf


class TrackerAPI(object):


    def __init__(self, chosen_video=None, **kwargs):

        self.chosen_video = chosen_video

        self.number_of_animals            = None # Number of animals
        self.accumulation_network_params  = None # Network params
        self.restoring_first_accumulation = False # Flag restores first accumulation
        self.accumulation_step_finished   = False # Flag accumulation step finished




    def start_tracking(self,
        one_animal_call=None,
        one_global_fragment_call=None,
        not_been_executed_call=None,
        post_processing_call=None,
        residual_identification_wo_ident_call=None,
        residual_identification_no_wo_ident_call=None,
        protocol3_accumulation_call=None,
        protocol3_pretraining_call=None,
        protocols1_and_2_call=None,
        not_protocols1_and_2_call=None
    ):


        if self.chosen_video.video.number_of_animals == 1:
            # GUI CALL
            if one_animal_call:
                one_animal_call()
            else:
                self.track_single_animal()

        elif self.chosen_video.list_of_global_fragments.number_of_global_fragments == 1:
            # GUI CALL
            if one_global_fragment_call:
                one_global_fragment_call()
            else:
                self.track_single_global_fragment_video()

        else:
            tf.reset_default_graph()
            self.chosen_video.video.accumulation_trial = 0
            delete = not self.chosen_video.processes_to_restore[
                'protocols1_and_2'] if 'protocols1_and_2' in self.chosen_video.processes_to_restore.keys() else True
            self.chosen_video.video.create_accumulation_folder(iteration_number=0, delete=delete)
            self.number_of_animals = self.chosen_video.video.number_of_animals if not self.chosen_video.video.identity_transfer \
                else self.chosen_video.video.knowledge_transfer_info_dict['number_of_animals']
            self.restoring_first_accumulation = False
            self.init_accumulation_network()
            self.chosen_video.video._tracking_with_knowledge_transfer = False if self.accumulation_network_params.knowledge_transfer_folder is None else True



            if not_been_executed_call: not_been_executed_call()



            if 'post_processing' in self.chosen_video.processes_to_restore and self.chosen_video.processes_to_restore[
                'post_processing']:
                self.restore_trajectories()
                self.restore_crossings_solved()
                self.restore_trajectories_wo_gaps()
                if post_processing_call: post_processing_call()

            elif 'residual_identification' in self.chosen_video.processes_to_restore and self.chosen_video.processes_to_restore[
                'residual_identification']:
                if self.chosen_video.video.track_wo_identities:
                    self.restore_trajectories()
                    if residual_identification_wo_ident_call:
                        residual_identification_wo_ident_call()

                else:
                    logger.info("Restoring residual identification")
                    self.restore_identification()
                    self.chosen_video.video._has_been_assigned = True

                    # GUI CALL
                    if residual_identification_no_wo_ident_call:
                        residual_identification_no_wo_ident_call()
                    else:
                        self.create_trajectories()

            elif 'protocol3_accumulation' in self.chosen_video.processes_to_restore and self.chosen_video.processes_to_restore[
                'protocol3_accumulation']:
                logger.info("Restoring second accumulation")
                self.restore_second_accumulation()
                self.chosen_video.video._first_frame_first_global_fragment = self.chosen_video.video._first_frame_first_global_fragment
                logger.warning(
                    'first_frame_first_global_fragment ' + str(self.chosen_video.video.first_frame_first_global_fragment))
                logger.info("Starting identification")

                #GUI CALL
                if protocol3_accumulation_call:
                    protocol3_accumulation_call()
                else:
                    self.create_trajectories()

            elif 'protocol3_pretraining' in self.chosen_video.processes_to_restore and self.chosen_video.processes_to_restore[
                'protocol3_pretraining']:
                logger.info("Restoring pretraining")
                logger.info("Initialising pretraining network")
                self.init_pretraining_net()
                logger.info("Restoring pretraining")
                self.accumulation_step_finished = True
                self.restore_first_accumulation()
                self.restore_pretraining()
                self.accumulation_manager.ratio_accumulated_images = self.chosen_video.video.percentage_of_accumulated_images[0]
                self.chosen_video.video._first_frame_first_global_fragment = [
                    self.chosen_video.video._first_frame_first_global_fragment[0]]
                self.chosen_video.video._percentage_of_accumulated_images = [
                    self.chosen_video.video.percentage_of_accumulated_images[0]]
                logger.info("Start accumulation parachute")
                # GUI CALL
                if protocol3_pretraining_call:
                    protocol3_pretraining_call()
                else:
                    self.accumulate()

            elif 'protocols1_and_2' in self.chosen_video.processes_to_restore and self.chosen_video.processes_to_restore[
                'protocols1_and_2']:
                logger.info("Restoring protocol 1 and 2")
                self.restoring_first_accumulation = True
                self.restore_first_accumulation()
                self.accumulation_manager.ratio_accumulated_images = self.chosen_video.video.percentage_of_accumulated_images[0]
                self.chosen_video.video._first_frame_first_global_fragment = [
                    self.chosen_video.video._first_frame_first_global_fragment[0]]
                self.chosen_video.video._percentage_of_accumulated_images = [
                    self.chosen_video.video.percentage_of_accumulated_images[0]]
                self.accumulation_step_finished = True
                # GUI CALL
                if protocols1_and_2_call:
                    protocols1_and_2_call()
                else:
                    self.accumulate()


            elif 'protocols1_and_2' not in self.chosen_video.processes_to_restore or not self.chosen_video.processes_to_restore[
                'protocols1_and_2']:
                logger.info("Starting protocol cascade")
                if not_protocols1_and_2_call:
                    not_protocols1_and_2_call()
                else:
                    self.protocol1()
                    # TODO: OR
                    #self.track_wo_identities()














































    def track_single_animal(self, create_trajectories=None):

        if create_trajectories is None:
            create_trajectories = self.create_trajectories

        logger.debug("------------------------> track_single_animal")
        [setattr(b, '_identity', 1) for bf in self.chosen_video.list_of_blobs.blobs_in_video for b in bf]
        [setattr(b, '_P2_vector', [1.]) for bf in self.chosen_video.list_of_blobs.blobs_in_video for b in bf]
        [setattr(b, 'frame_number', frame_number) for frame_number, bf in enumerate(self.chosen_video.list_of_blobs.blobs_in_video) for b in bf]
        create_trajectories()

    def track_single_global_fragment_video(self, create_trajectories=None):

        if create_trajectories is None:
            create_trajectories = self.create_trajectories

        logger.debug("------------------------> track_single_global_fragment_video")
        def get_P2_vector(identity, number_of_animals):
            P2_vector = np.zeros(number_of_animals)
            P2_vector[identity] = 1.
            return P2_vector
        [setattr(b, '_identity', b.fragment_identifier+1) for bf in self.chosen_video.list_of_blobs.blobs_in_video for b in bf]
        [setattr(b, '_P2_vector', get_P2_vector(b.fragment_identifier, self.chosen_video.video.number_of_animals))
            for bf in self.chosen_video.list_of_blobs.blobs_in_video for b in bf]
        self.chosen_video.video.accumulation_trial = 0
        self.chosen_video.video._first_frame_first_global_fragment = [0] # in case
        create_trajectories()

    def init_accumulation_network(self):
        self.accumulation_network_params = NetworkParams(self.number_of_animals,
            learning_rate = conf.LEARNING_RATE_IDCNN_ACCUMULATION,
            keep_prob = conf.KEEP_PROB_IDCNN_ACCUMULATION,
            scopes_layers_to_optimize = conf.LAYERS_TO_OPTIMISE_ACCUMULATION,
            save_folder = self.chosen_video.video.accumulation_folder,
            image_size = self.chosen_video.video.identification_image_size,
            video_path = self.chosen_video.video.video_path
        )

    def track_wo_identities(self, create_trajectories=None):

        if create_trajectories is None:
            create_trajectories = self.create_trajectories

        self.chosen_video.video.accumulation_trial = 0
        self.chosen_video.video._first_frame_first_global_fragment = [0]
        self.chosen_video.video._track_wo_identities = True
        create_trajectories()


    def protocol1(self, create_popup=None):
        logger.debug("****** setting protocol1 time")
        self.chosen_video.video._protocol1_time = time.time()
        self.chosen_video.list_of_fragments.reset(roll_back_to = 'fragmentation')
        self.chosen_video.list_of_global_fragments.reset(roll_back_to = 'fragmentation')
        self.net = ConvNetwork(self.accumulation_network_params)
        if self.chosen_video.video.identity_transfer:
            self.accumulation_network_params._restore_folder = self.chosen_video.video.knowledge_transfer_model_folder
        if self.chosen_video.video.tracking_with_knowledge_transfer:
            logger.debug('Restoring for knowledge transfer')
            tf.reset_default_graph()
            self.net = ConvNetwork(self.accumulation_network_params)
            self.net.restore()

        self.chosen_video.video._first_frame_first_global_fragment.append(self.chosen_video.list_of_global_fragments.set_first_global_fragment_for_accumulation(self.chosen_video.video, net=self.net, accumulation_trial=0))
        self.chosen_video.list_of_global_fragments.order_by_distance_to_the_first_global_fragment_for_accumulation(self.chosen_video.video, accumulation_trial = 0)
        self.accumulation_manager = AccumulationManager(self.chosen_video.video, self.chosen_video.list_of_fragments,
                                                        self.chosen_video.list_of_global_fragments,
                                                        threshold_acceptable_accumulation = conf.THRESHOLD_ACCEPTABLE_ACCUMULATION)
        self.global_step = 0

        if create_popup: create_popup()
        #self.create_one_shot_accumulation_popup()

        self.accumulation_step_finished = True
        self.accumulation_loop()


    def one_shot_accumulation(self, save_summaries=True, call_accumulate=True):
        logger.warning('Starting one_shot_accumulation')
        self.accumulation_step_finished = False
        self.accumulation_manager.ratio_accumulated_images,\
        self.store_validation_accuracy_and_loss_data,\
        self.store_training_accuracy_and_loss_data = perform_one_accumulation_step(
            self.accumulation_manager,
            self.chosen_video.video,
            self.global_step,
            self.net,
            save_summaries = save_summaries,
            GUI_axes = None,
            net_properties = None,
            plot_flag = conf.PLOT_ACCUMULATION_STEPS
        )
        self.accumulation_step_finished = True

        if call_accumulate:
            self.accumulate()



    def __accumulate_handler_call_accumulate(self):
        if self.accumulation_manager.continue_accumulation:
            self.accumulate()

    def __accumulate_handler_identification_popup_open(self):
        self.identify()
        self.postprocess_impossible_jumps()


    def accumulate(self,
            identification_popup_open           = None,
            one_shot_accumulation_popup_dismiss = None,
            create_pretraining_popup            = None,
            unschedule_accumulate               = None,
            call_accumulate                     = True
        ):

        if identification_popup_open is None:
            identification_popup_open = self.__accumulate_handler_identification_popup_open

        logger.info("------------------------> Calling accumulate")

        if self.accumulation_step_finished and self.accumulation_manager.continue_accumulation:

            logger.info("--------------------> Performing accumulation")
            if self.accumulation_manager.counter == 1 and self.chosen_video.video.accumulation_trial == 0:
                self.chosen_video.video._protocol1_time = time.time()-self.chosen_video.video.protocol1_time
                self.chosen_video.video._protocol2_time = time.time()

            self.one_shot_accumulation()
            call_accumulate = False

        elif not self.accumulation_manager.continue_accumulation\
            and not self.chosen_video.video.first_accumulation_finished\
            and self.accumulation_manager.ratio_accumulated_images > conf.THRESHOLD_EARLY_STOP_ACCUMULATION:

            logger.info("Protocol 1 successful")
            self.save_after_first_accumulation()
            if 'protocols1_and_2' not in self.chosen_video.processes_to_restore or not self.chosen_video.processes_to_restore['protocols1_and_2']:
                self.chosen_video.video._protocol1_time = time.time()-self.chosen_video.video.protocol1_time

            # call handler
            identification_popup_open()

        elif not self.accumulation_manager.continue_accumulation\
            and not self.chosen_video.video.has_been_pretrained:

            self.save_after_first_accumulation()

            if self.accumulation_manager.ratio_accumulated_images > conf.THRESHOLD_ACCEPTABLE_ACCUMULATION:
                logger.info("Protocol 2 successful")
                logger.warning("------------------------ dismissing one shot accumulation popup")
                if one_shot_accumulation_popup_dismiss: one_shot_accumulation_popup_dismiss() # UPDATE GUI

                self.save_after_first_accumulation()
                if 'protocols1_and_2' not in self.chosen_video.processes_to_restore or not self.chosen_video.processes_to_restore['protocols1_and_2']:
                    self.chosen_video.video._protocol2_time = time.time()-self.chosen_video.video.protocol2_time

                # call handler
                identification_popup_open()

            elif self.accumulation_manager.ratio_accumulated_images < conf.THRESHOLD_ACCEPTABLE_ACCUMULATION:

                logger.info("Protocol 2 failed -> Start protocol 3")
                if 'protocols1_and_2' not in self.chosen_video.processes_to_restore or not self.chosen_video.processes_to_restore['protocols1_and_2']:
                    self.chosen_video.video._protocol1_time = time.time()-self.chosen_video.video.protocol1_time
                    if self.chosen_video.video.protocol2_time != 0:
                        self.chosen_video.video._protocol2_time = time.time()-self.chosen_video.video.protocol2_time
                self.chosen_video.video._protocol3_pretraining_time = time.time()

                self.pretraining_counter = 0

                # call handler
                if create_pretraining_popup:
                    create_pretraining_popup()

                self.protocol3()

        elif self.chosen_video.video.has_been_pretrained\
            and self.chosen_video.video.accumulation_trial < conf.MAXIMUM_NUMBER_OF_PARACHUTE_ACCUMULATIONS\
            and self.accumulation_manager.ratio_accumulated_images < conf.THRESHOLD_ACCEPTABLE_ACCUMULATION:

            logger.info("Accumulation in protocol 3 is not successful. Opening parachute ...")
            if self.chosen_video.video.accumulation_trial == 0:
                self.chosen_video.video._protocol3_accumulation_time = time.time()
            self.chosen_video.video.accumulation_trial += 1
            if not self.accumulation_manager.continue_accumulation and self.chosen_video.video.accumulation_trial > 1:
                self.save_and_update_accumulation_parameters_in_parachute()
            self.accumulation_parachute_init(self.chosen_video.video.accumulation_trial)

            call_accumulate = False
            self.accumulation_loop()


        elif self.chosen_video.video.has_been_pretrained and\
            (self.accumulation_manager.ratio_accumulated_images >= conf.THRESHOLD_ACCEPTABLE_ACCUMULATION\
            or self.chosen_video.video.accumulation_trial >= conf.MAXIMUM_NUMBER_OF_PARACHUTE_ACCUMULATIONS):

            logger.info("Accumulation after protocol 3 has been successful")
            if 'protocol3_accumulation' not in self.chosen_video.processes_to_restore:
                self.chosen_video.video._protocol3_accumulation_time = time.time()-self.chosen_video.video.protocol3_accumulation_time
            elif 'protocol3_accumulation' in self.chosen_video.processes_to_restore and not self.chosen_video.processes_to_restore['protocol3_accumulation']:
                self.chosen_video.video._protocol3_accumulation_time = time.time()-self.chosen_video.video.protocol3_accumulation_time
            else:
                self.chosen_video.video._protocol3_accumulation_time = time.time()-self.chosen_video.video.protocol3_accumulation_time

            logger.warning("************************ Unscheduling accumulate")
            # call handler
            if unschedule_accumulate:
                unschedule_accumulate()

            logger.warning("------------------------ dismissing one shot accumulation popup")
            # call handler
            if one_shot_accumulation_popup_dismiss:
                one_shot_accumulation_popup_dismiss()

            self.save_after_second_accumulation()
            logger.info("Start residual indentification")
            # call handler
            identification_popup_open()

        # call handler
        if call_accumulate:
            self.__accumulate_handler_call_accumulate()


    def save_and_update_accumulation_parameters_in_parachute(self):
        logger.warning("self.accumulation_manager.ratio_accumulated_images %.4f" %self.accumulation_manager.ratio_accumulated_images)
        self.chosen_video.video._ratio_accumulated_images = self.accumulation_manager.ratio_accumulated_images
        self.chosen_video.video._percentage_of_accumulated_images.append(self.chosen_video.video.ratio_accumulated_images)
        self.chosen_video.list_of_fragments.save_light_list(self.chosen_video.video._accumulation_folder)




    def accumulation_loop(self, do_accumulate=True):
        logger.warning('------------Calling accumulation loop')
        self.chosen_video.video.init_accumulation_statistics_attributes()
        self.accumulation_manager.threshold_early_stop_accumulation = conf.THRESHOLD_EARLY_STOP_ACCUMULATION
        logger.warning('Calling accumulate from accumulation_loop')

        if do_accumulate: self.accumulate()


    def accumulation_parachute_init(self,
        iteration_number,
        one_shot_accumulation_popup_dismiss=None):
        logger.debug("------------------------> accumulation_parachute_init")
        logger.info("Starting accumulation %i" %iteration_number)

        # Call to GUI
        if one_shot_accumulation_popup_dismiss:
            one_shot_accumulation_popup_dismiss()

        delete = not self.chosen_video.processes_to_restore['protocol3_accumulation'] if 'protocol3_accumulation' in self.chosen_video.processes_to_restore.keys() else True
        self.chosen_video.video.create_accumulation_folder(iteration_number = iteration_number, delete = delete)
        self.chosen_video.video.accumulation_trial = iteration_number
        self.chosen_video.list_of_fragments.reset(roll_back_to = 'fragmentation')
        self.chosen_video.list_of_global_fragments.reset(roll_back_to = 'fragmentation')
        if self.chosen_video.video.identity_transfer:
            tf.reset_default_graph()
            self.accumulation_network_params._restore_folder = self.chosen_video.video.knowledge_transfer_model_folder
            self.net = ConvNetwork(self.accumulation_network_params)
            self.net.restore()
        self.chosen_video.video._first_frame_first_global_fragment.append(self.chosen_video.list_of_global_fragments.set_first_global_fragment_for_accumulation(self.chosen_video.video, net=self.net, accumulation_trial = iteration_number - 1))
        self.chosen_video.list_of_global_fragments.order_by_distance_to_the_first_global_fragment_for_accumulation(self.chosen_video.video, accumulation_trial = iteration_number - 1)
        logger.warning('first_frame_first_global_fragment ' + str(self.chosen_video.video.first_frame_first_global_fragment))
        logger.info("We will restore the network from a previous pretraining: %s" %self.chosen_video.video.pretraining_folder)
        self.accumulation_network_params.save_folder = self.chosen_video.video.accumulation_folder
        self.accumulation_network_params._restore_folder = self.chosen_video.video.pretraining_folder
        self.accumulation_network_params.scopes_layers_to_optimize = ['fully-connected1','fully_connected_pre_softmax']
        logger.info("Initialising accumulation network")
        tf.reset_default_graph()
        self.net = ConvNetwork(self.accumulation_network_params)
        self.net.restore()
        self.net.reinitialize_softmax_and_fully_connected()
        logger.info("Initialising accumulation manager")
        self.accumulation_manager = AccumulationManager(self.chosen_video.video,
                                                    self.chosen_video.list_of_fragments, self.chosen_video.list_of_global_fragments,
                                                    threshold_acceptable_accumulation = conf.THRESHOLD_ACCEPTABLE_ACCUMULATION)
        logger.info("Start accumulation")
        self.global_step = 0


    def save_after_first_accumulation(self):
        """Set flags and save data"""
        logger.info("Saving first accumulation paramters")

        if not self.restoring_first_accumulation:
            self.chosen_video.video._first_accumulation_finished = True
            self.chosen_video.video._ratio_accumulated_images = self.accumulation_manager.ratio_accumulated_images
            self.chosen_video.video._percentage_of_accumulated_images = [self.chosen_video.video.ratio_accumulated_images]
            self.chosen_video.video._accumulation_network_params = self.accumulation_network_params
            self.chosen_video.video.save()
            self.chosen_video.list_of_fragments.save(self.chosen_video.video.fragments_path)
            self.chosen_video.list_of_global_fragments.save(self.chosen_video.video.global_fragments_path, self.chosen_video.list_of_fragments.fragments)
            self.chosen_video.list_of_fragments.save_light_list(self.chosen_video.video._accumulation_folder)



    def save_after_second_accumulation(self):
        logger.info("Saving second accumulation parameters")
        self.save_and_update_accumulation_parameters_in_parachute()
        self.chosen_video.video.accumulation_trial = np.argmax(self.chosen_video.video.percentage_of_accumulated_images)
        self.chosen_video.video._ratio_accumulated_images = self.chosen_video.video.percentage_of_accumulated_images[self.chosen_video.video.accumulation_trial]
        accumulation_folder_name = 'accumulation_' + str(self.chosen_video.video.accumulation_trial)
        self.chosen_video.video._accumulation_folder = os.path.join(self.chosen_video.video.session_folder, accumulation_folder_name)
        self.chosen_video.list_of_fragments.load_light_list(self.chosen_video.video._accumulation_folder)
        self.chosen_video.video._second_accumulation_finished = True
        logger.info("Saving global fragments")
        self.chosen_video.list_of_fragments.save(self.chosen_video.video.fragments_path)
        self.chosen_video.list_of_global_fragments.save(self.chosen_video.video.global_fragments_path, self.chosen_video.list_of_fragments.fragments)
        logger.info("Restoring networks to best second accumulation")
        self.accumulation_network_params.restore_folder = self.chosen_video.video._accumulation_folder
        self.net = ConvNetwork(self.accumulation_network_params)
        self.net.restore()
        self.chosen_video.video._accumulation_network_params = self.accumulation_network_params
        self.chosen_video.video.save()


    def init_pretraining_variables(self):
        self.init_pretraining_net()
        self.pretraining_global_step = 0
        self.net = ConvNetwork(self.pretrain_network_params)
        self.ratio_of_pretrained_images = 0
        if self.chosen_video.video.tracking_with_knowledge_transfer:
            self.net.restore()
        self.store_training_accuracy_and_loss_data_pretrain = Store_Accuracy_and_Loss(self.net,
                                                                                    name = 'training',
                                                                                    scope = 'pretraining')
        self.store_validation_accuracy_and_loss_data_pretrain = Store_Accuracy_and_Loss(self.net,
                                                                                    name = 'validation',
                                                                                    scope = 'pretraining')

    def pretraining_loop(self, call_from_gui=False):
        self.chosen_video.list_of_fragments.reset(roll_back_to = 'fragmentation')
        self.chosen_video.list_of_global_fragments.order_by_distance_travelled()

        ## IT SHOULD NOT BE CALLED BY THE GUI
        if not call_from_gui:
            self.one_shot_pretraining()
            self.continue_pretraining()

    def continue_pretraining(self, clock_unschedule=None):
        if self.pretraining_step_finished and self.ratio_of_pretrained_images < conf.MAX_RATIO_OF_PRETRAINED_IMAGES:
            self.one_shot_pretraining()

            if clock_unschedule is None:
                self.continue_pretraining()

        elif self.ratio_of_pretrained_images > conf.MAX_RATIO_OF_PRETRAINED_IMAGES:
            self.chosen_video.video._has_been_pretrained = True

            # Call GUI
            if clock_unschedule: clock_unschedule()

            logger.warning('Calling accumulate from continue_pretraining')
            logger.debug('****** saving protocol3 pretraining time')
            self.chosen_video.video._protocol3_pretraining_time = time.time()-self.chosen_video.video.protocol3_pretraining_time
            self.accumulate()

    def one_shot_pretraining(self, generate_tensorboard=False, gui_graph_canvas=None):
        self.pretraining_step_finished = False
        self.pretraining_global_fragment = self.chosen_video.list_of_global_fragments.global_fragments[self.pretraining_counter]
        self.net,\
        self.ratio_of_pretrained_images,\
        pretraining_global_step,\
        self.store_training_accuracy_and_loss_data_pretrain,\
        self.store_validation_accuracy_and_loss_data_pretrain,\
        self.chosen_video.list_of_fragments = pre_train_global_fragment(self.net,
                                                    self.pretraining_global_fragment,
                                                    self.chosen_video.list_of_fragments,
                                                    self.pretraining_global_step,
                                                    True, True,
                                                    generate_tensorboard,
                                                    self.store_training_accuracy_and_loss_data_pretrain,
                                                    self.store_validation_accuracy_and_loss_data_pretrain,
                                                    print_flag = False,
                                                    plot_flag = False,
                                                    batch_size = conf.BATCH_SIZE_IDCNN,
                                                    canvas_from_GUI = gui_graph_canvas)
        self.pretraining_counter += 1
        self.pretraining_step_finished = True

    def protocol3(self):

        self.init_pretraining_variables()
        number_of_images_in_global_fragments = self.chosen_video.video.number_of_unique_images_in_global_fragments
        if self.chosen_video.old_video and self.chosen_video.old_video.first_accumulation_finished == True:
            self.chosen_video.list_of_global_fragments.reset(roll_back_to = 'fragmentation')
            self.chosen_video.list_of_fragments.reset(roll_back_to = 'fragmentation')
        logger.info("Starting pretraining. Checkpoints will be stored in %s" %self.chosen_video.video.pretraining_folder)
        if self.chosen_video.video.tracking_with_knowledge_transfer:
            logger.info("Performing knowledge transfer from %s" %self.chosen_video.video.knowledge_transfer_model_folder)
            self.pretrain_network_params.knowledge_transfer_folder = self.chosen_video.video.knowledge_transfer_model_folder
        logger.info("Start pretraining")
        self.pretraining_step_finished = True
        self.pretraining_loop()






    def init_pretraining_net(self):
        delete = not self.chosen_video.processes_to_restore['protocol3_pretraining'] if 'protocol3_pretraining' in self.chosen_video.processes_to_restore.keys() else True
        self.chosen_video.video.create_pretraining_folder(delete = delete)
        self.pretrain_network_params = NetworkParams(self.chosen_video.video.number_of_animals,
                                                learning_rate = conf.LEARNING_RATE_IDCNN_PRETRAINING,
                                                keep_prob = conf.KEEP_PROB_IDCNN_PRETRAINING,
                                                scopes_layers_to_optimize = conf.LAYERS_TO_OPTIMISE_PRETRAINING,
                                                save_folder = self.chosen_video.video.pretraining_folder,
                                                image_size = self.chosen_video.video.identification_image_size,
                                                video_path = self.chosen_video.video.video_path)
        self.chosen_video.video._pretraining_network_params = self.pretrain_network_params



    def restore_video_attributes(self):
        list_of_attributes = ['accumulation_folder',
                    'second_accumulation_finished',
                    'number_of_accumulated_global_fragments',
                    'number_of_non_certain_global_fragments',
                    'number_of_randomly_assigned_global_fragments',
                    'number_of_nonconsistent_global_fragments',
                    'number_of_nonunique_global_fragments',
                    'number_of_acceptable_global_fragments',
                    'validation_accuracy', 'validation_individual_accuracies',
                    'training_accuracy', 'training_individual_accuracies',
                    'percentage_of_accumulated_images', 'accumulation_trial',
                    'ratio_accumulated_images', 'first_accumulation_finished',
                    'identity_transfer', 'accumulation_statistics',
                    'first_frame_first_global_fragment', 'pretraining_folder',
                    'has_been_pretrained', 'has_been_assigned',
                    'has_crossings_solved','has_trajectories',
                    'has_trajectories_wo_gaps',
                    'protocol1_time', 'protocol2_time',
                    'protocol3_pretraining_time', 'protocol3_accumulation_time',
                    'identify_time', 'create_trajectories_time']
        is_property = [True, True, False, False, False, False, False, False,
                        False, False, False, False, True, False, True, True,
                        True, False, True, True, True, True, True, True, True,
                        True, True, True, True, True, True]
        self.chosen_video.video.copy_attributes_between_two_video_objects(self.chosen_video.old_video, list_of_attributes, is_property = is_property)

    def restore_first_accumulation(self):
        self.restore_video_attributes()
        self.chosen_video.video._ratio_accumulated_images = self.chosen_video.video.percentage_of_accumulated_images[0]
        self.accumulation_network_params.restore_folder = self.chosen_video.video._accumulation_folder
        self.accumulation_manager = AccumulationManager(self.chosen_video.video, self.chosen_video.list_of_fragments,
                                                    self.chosen_video.list_of_global_fragments,
                                                    threshold_acceptable_accumulation = conf.THRESHOLD_ACCEPTABLE_ACCUMULATION)
        self.net = ConvNetwork(self.accumulation_network_params)
        self.net.restore()
        logger.info("Saving video")
        self.chosen_video.video._has_been_pretrained = False
        self.chosen_video.video.save()
        self.chosen_video.list_of_fragments.save_light_list(self.chosen_video.video._accumulation_folder)

    def restore_pretraining(self):
        logger.info("Restoring pretrained network")
        self.restore_video_attributes()
        self.pretrain_network_params.restore_folder = self.chosen_video.video.pretraining_folder
        self.net = ConvNetwork(self.pretrain_network_params)
        self.net.restore()
        self.accumulation_manager = AccumulationManager(self.chosen_video.video, self.chosen_video.list_of_fragments,
                                                    self.chosen_video.list_of_global_fragments,
                                                    threshold_acceptable_accumulation = conf.THRESHOLD_ACCEPTABLE_ACCUMULATION)
        self.chosen_video.video.accumulation_trial = 0
        self.chosen_video.video.save()

    def restore_second_accumulation(self):
        self.restore_video_attributes()
        logger.info("Restoring trained network")
        self.accumulation_network_params.restore_folder = self.chosen_video.video._accumulation_folder
        self.chosen_video.list_of_fragments.load_light_list(self.chosen_video.video._accumulation_folder)
        self.net = ConvNetwork(self.accumulation_network_params)
        self.net.restore()
        self.chosen_video.video.save()

    def restore_identification(self):
        self.restore_video_attributes()
        self.chosen_video.list_of_fragments.load_light_list(self.chosen_video.video._accumulation_folder)
        self.chosen_video.video.save()

    def restore_trajectories(self):
        self.restore_video_attributes()
        self.chosen_video.video.save()

    def restore_crossings_solved(self):
        self.restore_video_attributes()
        self.chosen_video.video.copy_attributes_between_two_video_objects(self.chosen_video.old_video, ['blobs_no_gaps_path'], [False])
        self.chosen_video.list_of_blobs_no_gaps = ListOfBlobs.load(self.chosen_video.video, self.chosen_video.video.blobs_no_gaps_path)
        self.chosen_video.video.save()

    def restore_trajectories_wo_gaps(self):
        self.restore_video_attributes()
        self.chosen_video.video.save()

    def identify(self):
        self.chosen_video.video._identify_time = time.time()
        logger.warning("In identify")
        self.chosen_video.list_of_fragments.reset(roll_back_to = 'accumulation')
        logger.warning("Assigning remaining fragments")
        assign_remaining_fragments(self.chosen_video.list_of_fragments, self.chosen_video.video, self.net)
        self.chosen_video.video._has_been_assigned = True
        self.chosen_video.video.save()


    def postprocess_impossible_jumps(self, call_update_list_of_blobs=True):
        if not hasattr(self.chosen_video.video, 'velocity_threshold') and hasattr(self.chosen_video.old_video,'velocity_threshold'):
            self.chosen_video.video.velocity_threshold = self.chosen_video.old_video.velocity_threshold
        elif not hasattr(self.chosen_video.old_video, 'velocity_threshold'):
            self.chosen_video.video.velocity_threshold = compute_model_velocity(
                                                                self.chosen_video.list_of_fragments.fragments,
                                                                self.chosen_video.video.number_of_animals,
                                                                percentile = conf.VEL_PERCENTILE)
        correct_impossible_velocity_jumps(self.chosen_video.video, self.chosen_video.list_of_fragments)
        self.chosen_video.list_of_fragments.save(self.chosen_video.video.fragments_path)
        self.chosen_video.video.save()

        if call_update_list_of_blobs:
            self.update_list_of_blobs()


    def update_list_of_blobs(self, create_trajectories=None):

        if create_trajectories is None:
            create_trajectories = self.create_trajectories

        self.chosen_video.video.individual_fragments_stats = self.chosen_video.list_of_fragments.get_stats(self.chosen_video.list_of_global_fragments)
        self.chosen_video.video.compute_overall_P2(self.chosen_video.list_of_fragments.fragments)
        self.chosen_video.list_of_fragments.save_light_list(self.chosen_video.video._accumulation_folder)
        self.chosen_video.video.save()
        if not hasattr(self.chosen_video, 'list_of_blobs'):
            self.chosen_video.list_of_blobs = ListOfBlobs.load(self.chosen_video.video, self.chosen_video.old_video.blobs_path)
        self.chosen_video.list_of_blobs.update_from_list_of_fragments(self.chosen_video.list_of_fragments.fragments,
                                                    self.chosen_video.video.fragment_identifier_to_index)
        # if False:
        #     self.chosen_video.list_of_blobs.compute_nose_and_head_coordinates()
        self.chosen_video.list_of_blobs.save(self.chosen_video.video,
                                        self.chosen_video.video.blobs_path,
                                        number_of_chunks = self.chosen_video.video.number_of_frames)
        self.chosen_video.video._identify_time = time.time()-self.chosen_video.video.identify_time
        create_trajectories()

    def create_trajectories(self,
        trajectories_popup_dismiss=None,
        interpolate_crossings=None,
        update_and_show_happy_ending_popup=None
    ):

        if interpolate_crossings is None:
            interpolate_crossings = self.interpolate_crossings


        self.chosen_video.video._create_trajectories_time = time.time()
        if 'post_processing' not in self.chosen_video.processes_to_restore or not self.chosen_video.processes_to_restore['post_processing']:
            if not self.chosen_video.video.track_wo_identities:
                self.chosen_video.video.create_trajectories_folder()
                trajectories_file = os.path.join(self.chosen_video.video.trajectories_folder, 'trajectories.npy')
                trajectories = produce_output_dict(self.chosen_video.list_of_blobs.blobs_in_video, self.chosen_video.video)
            else:
                self.chosen_video.video.create_trajectories_wo_identities_folder()
                trajectories_file = os.path.join(self.chosen_video.video.trajectories_wo_identities_folder, 'trajectories_wo_identities.npy')
                trajectories = produce_output_dict(self.chosen_video.list_of_blobs.blobs_in_video, self.chosen_video.video)
            np.save(trajectories_file, trajectories)
            logger.info("Saving trajectories")
            np.save(trajectories_file, trajectories)
        self.chosen_video.video._has_trajectories = True

        # Call GUI function
        if trajectories_popup_dismiss: trajectories_popup_dismiss()


        if self.chosen_video.video.number_of_animals != 1 and self.chosen_video.list_of_global_fragments.number_of_global_fragments != 1 and not self.chosen_video.video.track_wo_identities:
            # Call GUI function
            interpolate_crossings()
        else:
            self.chosen_video.video.overall_P2 = 1.
            self.chosen_video.video._has_been_assigned = True
            self.chosen_video.video._has_crossings_solved = False
            self.chosen_video.video._has_trajectories_wo_gaps = False
            self.chosen_video.list_of_blobs.save(self.chosen_video.video,
                                            self.chosen_video.video.blobs_path,
                                            number_of_chunks = self.chosen_video.video.number_of_frames)
            # Call GUI function
            if update_and_show_happy_ending_popup: update_and_show_happy_ending_popup()
        self.chosen_video.video.save()



    def interpolate_crossings(self, interpolate_crossings_popups_actions=None):

        if interpolate_crossings_popups_actions is None:
            interpolate_crossings_popups_actions = self.create_trajectories_wo_gaps

        self.chosen_video.list_of_blobs_no_gaps = copy.deepcopy(self.chosen_video.list_of_blobs)
        self.chosen_video.video._has_crossings_solved = False
        self.chosen_video.list_of_blobs_no_gaps = close_trajectories_gaps(self.chosen_video.video, self.chosen_video.list_of_blobs_no_gaps, self.chosen_video.list_of_fragments)
        self.chosen_video.video.blobs_no_gaps_path = os.path.join(os.path.split(self.chosen_video.video.blobs_path)[0], 'blobs_collection_no_gaps.npy')
        self.chosen_video.list_of_blobs_no_gaps.save(self.chosen_video.video, path_to_save = self.chosen_video.video.blobs_no_gaps_path, number_of_chunks = self.chosen_video.video.number_of_frames)
        self.chosen_video.video._has_crossings_solved = True
        self.chosen_video.video.save()

        interpolate_crossings_popups_actions()



    def create_trajectories_wo_gaps(self):
        self.chosen_video.video.create_trajectories_wo_gaps_folder()
        logger.info("Generating trajectories. The trajectories files are stored in %s" %self.chosen_video.video.trajectories_wo_gaps_folder)
        trajectories_wo_gaps_file = os.path.join(self.chosen_video.video.trajectories_wo_gaps_folder, 'trajectories_wo_gaps.npy')
        trajectories_wo_gaps = produce_output_dict(self.chosen_video.list_of_blobs_no_gaps.blobs_in_video, self.chosen_video.video)
        np.save(trajectories_wo_gaps_file, trajectories_wo_gaps)
        self.chosen_video.video._has_trajectories_wo_gaps = True
        logger.info("Saving trajectories")
        self.chosen_video.list_of_blobs = assign_zeros_with_interpolation_identities(self.chosen_video.list_of_blobs, self.chosen_video.list_of_blobs_no_gaps)
        trajectories_file = os.path.join(self.chosen_video.video.trajectories_folder, 'trajectories.npy')
        trajectories = produce_output_dict(self.chosen_video.list_of_blobs.blobs_in_video, self.chosen_video.video)
        np.save(trajectories_file, trajectories)
        self.chosen_video.video.save()
        self.chosen_video.video._create_trajectories_time = time.time()-self.chosen_video.video.create_trajectories_time

    def update_and_show_happy_ending_popup(self):
        if not hasattr(self.chosen_video.video, 'overall_P2'):
            self.chosen_video.video.compute_overall_P2(self.chosen_video.list_of_fragments.fragments)
        self.chosen_video.video.save()

    def network_params_to_string(self):
        self.str_model = str(self.accumulation_network_params.cnn_model)
        self.str_lr = str(self.accumulation_network_params.learning_rate)
        self.str_kp = str(self.accumulation_network_params.keep_prob)
        self.str_optimiser = "SGD" if not self.accumulation_network_params.use_adam_optimiser else "Adam"
        self.str_layers_to_train = "all" if self.accumulation_network_params.scopes_layers_to_optimize is None else str(self.accumulation_network_params.scopes_layers_to_optimize)
        self.restore_folder = self.accumulation_network_params.restore_folder if self.accumulation_network_params.restore_folder is not None else 'None'
        self.save_folder = self.accumulation_network_params.save_folder if self.accumulation_network_params.save_folder is not None else 'None'
        self.knowledge_transfer_folder = self.accumulation_network_params.knowledge_transfer_folder if self.accumulation_network_params.knowledge_transfer_folder is not None else 'None'
        # self.kt_conv_layers_to_discard = self.accumulation_network_params.kt_conv_layers_to_discard if self.accumulation_network_params.kt_conv_layers_to_discard is not None else 'None'



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import os
import sys
import numpy as np
from idtrackerai.list_of_blobs import ListOfBlobs
from idtrackerai.list_of_fragments import ListOfFragments
from idtrackerai.blob import Blob
# from idtrackerai.utils.GUI_utils import selectDir, getInput
if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.generate_light_groundtruth_blob_list")

class GroundTruthBlob(object):
    """Lighter blob objects.
    Attributes:
        identity (preferring the one assigned by the user, if it is not None)
        centroid
        pixels (pixels is stored to check the groundtruth in crossings)
    """
    def __init__(self, attributes_to_get = ['identity', 'assigned_identities',
                                            'used_for_training', 'accumulation_step',
                                            'centroid', 'pixels',
                                            'frame_number',
                                            'is_an_individual', 'is_a_crossing',
                                            'was_a_crossing',
                                            'blob_index', 'fragment_identifier']):
        self.attributes = attributes_to_get

    def get_attribute(self, blob):
        for attribute in self.attributes:
            if attribute == 'identity':
                setattr(self, attribute, getattr(blob, 'final_identities')[0])
            if attribute == 'assigned_identities':
                setattr(self, attribute, getattr(blob, 'assigned_identities')[0])
            else:
                setattr(self, attribute, getattr(blob, attribute))

class GroundTruth(object):
    def __init__(self, video = [], blobs_in_video = [], start = None, end = None):
        self.video = video
        self.blobs_in_video = blobs_in_video
        self.start = start
        self.end = end

    def save(self, name = ''):
        if name == '':
            path_to_save_groundtruth = os.path.join(os.path.split(self.video.video_path)[0], '_groundtruth.npy')
        else:
            path_to_save_groundtruth = os.path.join(os.path.split(self.video.video_path)[0], '_groundtruth_' + name + '.npy')
        logger.info("saving ground truth at %s" %path_to_save_groundtruth)
        np.save(path_to_save_groundtruth, self)
        logger.info("done")

def generate_groundtruth(video, blobs_in_video = None, start = None, end = None,
                        wrong_crossing_counter = None, unidentified_individuals_counter = None,
                        save_gt = True):
    """Generates a list of light blobs_in_video, given a video object corresponding to a
    tracked video
    """
    #make sure the video has been succesfully tracked
    assert video.has_been_assigned == True
    blobs_in_video_groundtruth = []

    for blobs_in_frame in blobs_in_video:
        blobs_in_frame_groundtruth = []

        for blob in blobs_in_frame:
            gt_blob = GroundTruthBlob()
            gt_blob.get_attribute(blob)
            blobs_in_frame_groundtruth.append(gt_blob)

        blobs_in_video_groundtruth.append(blobs_in_frame_groundtruth)

    groundtruth = GroundTruth(video = video,
                            blobs_in_video = blobs_in_video_groundtruth,
                            start = start,
                            end = end)
    groundtruth.wrong_crossing_counter = wrong_crossing_counter
    groundtruth.unidentified_individuals_counter = unidentified_individuals_counter
    if save_gt:
        groundtruth.save()
    return groundtruth

if __name__ == "__main__":

    # session_path = selectDir('./') #select path to video
    video_path = os.path.join(session_path,'video_object.npy')
    video = np.load(video_path, allow_pickle=True).item()
    start = getInput('GroundTruth (start)', 'Input the starting frame for the interval in which the video has been validated')
    end = getInput('GroundTruth (end)', 'Input the ending frame for the interval in which the video has been validated')
    #read blob list from video
    # fragments_path = os.path.join(session_path, 'preprocessing', 'fragments.npy')
    # blobs_path = os.path.join(session_path, 'preprocessing', 'blobs_collection.npy')
    list_of_fragments = ListOfFragments.load(video.fragments_path)
    list_of_blobs = ListOfBlobs.load(video, video.blobs_path)
    list_of_blobs.update_from_list_of_fragments(list_of_fragments.fragments, video.fragment_identifier_to_index)
    groundtruth = generate_groundtruth(video, list_of_blobs.blobs_in_video, int(start), int(end))



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import sys

import cv2
import numpy as np
from confapp import conf

from idtrackerai.network.identification_model.get_data import duplicate_PCA_images

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.get_crossing_data_set")

class CrossingDataset(object):
    def __init__(self, blobs_list, video,
                crossings = [], individual_blobs = [], test = [],
                image_size = None, scope = '', dataset_image_size = 40):
        self.video = video
        self.scope = scope
        self.dataset_image_size = dataset_image_size
        self.blobs = blobs_list
        self.there_are_crossings = True
        self.get_video_height_and_width_according_to_resolution_reduction()
        if (scope == 'training' or scope == 'validation'):
            self.get_list_of_crossing_blobs_for_training(crossings, image_size)
            if self.there_are_crossings:
                self.get_list_of_individual_blobs_for_training(individual_blobs)
                logger.info("number of sure crossing images used for training: %i" %len(self.crossing_blobs))
                logger.info("number of individual images used for training: %i" %len(self.individual_blobs))
        if scope == 'test':
            self.image_size = image_size
            self.get_list_of_blobs_for_test(test)
            logger.info("number of test images: %i" %len(self.test))


    def get_video_height_and_width_according_to_resolution_reduction(self):
        if self.video.resolution_reduction == 1:
            self.video_height = self.video.height
            self.video_width = self.video.width
        else:
            self.video_height  = int(self.video.height * self.video.resolution_reduction)
            self.video_width  = int(self.video.width * self.video.resolution_reduction)

    def get_list_of_individual_blobs_for_training(self, individual_blobs):
        if len(individual_blobs) == 0:
            self.individual_blobs = [blob for blobs_in_frame in self.blobs for blob in blobs_in_frame
                                    if blob.is_a_sure_individual()
                                    or blob.in_a_global_fragment_core(blobs_in_frame)]
            logger.debug("number of individual blobs (before cut): %i" %len(self.individual_blobs))
            np.random.shuffle(self.individual_blobs)
            ratio = 1
            if len(self.individual_blobs) > ratio * len(self.crossing_blobs):
                self.individual_blobs = self.individual_blobs[:ratio * len(self.crossing_blobs)]

        else:
            self.individual_blobs = individual_blobs

    def get_list_of_crossing_blobs_for_training(self, crossings, image_size):
        if len(crossings) == 0 or image_size is None:
            self.crossing_blobs = [blob for blobs_in_frame in self.blobs for blob in blobs_in_frame
                                    if blob.is_a_sure_crossing()]
            logger.debug("number of crossing blobs (in get list): %i" %len(self.crossing_blobs))
            if len(self.crossing_blobs) <= conf.MINIMUM_NUMBER_OF_CROSSINGS_TO_TRAIN_CROSSING_DETECTOR:
                logger.debug("Not enough crossings where found")
                self.there_are_crossings = False
            else:
                self.there_are_crossings = True
                np.random.seed(0)
                np.random.shuffle(self.crossing_blobs)
                self.image_size = np.max([np.max(crossing.bounding_box_image.shape) for crossing in self.crossing_blobs]) + 5
        else:
            self.crossing_blobs = crossings
            self.image_size = image_size

    def get_list_of_blobs_for_test(self, test):
        if len(test) == 0:
            self.test = [blob for blobs_in_frame in self.blobs for blob in blobs_in_frame
                            if (blob.is_an_individual
                                and not blob.in_a_global_fragment_core(blobs_in_frame)
                                and not blob.is_a_sure_individual())
                            or (blob.is_a_crossing
                                and not blob.is_a_sure_crossing())]
        else:
            self.test = test

    def get_data(self, sampling_ratio_start = 0, sampling_ratio_end = 1.):
        logger.info("Generating crossing %s set." %self.scope)
        self.crossings_sliced = self.slice(self.crossing_blobs, sampling_ratio_start, sampling_ratio_end)
        self.crossings_images =  self.generate_crossing_images()
        self.crossing_labels = np.ones(len(self.crossings_images))
        if self.scope == "training":
            self.crossings_images, self.crossing_labels = duplicate_PCA_images(self.crossings_images, self.crossing_labels)
        assert len(self.crossing_labels) == len(self.crossings_images)
        logger.info("Generating single individual %s set" %self.scope)
        self.individual_blobs_sliced = self.slice(self.individual_blobs, sampling_ratio_start, sampling_ratio_end)
        self.individual_blobs_images = self.generate_individual_blobs_images()
        self.individual_blobs_labels = np.zeros(len(self.individual_blobs_images))
        assert len(self.individual_blobs_labels) == len(self.individual_blobs_images)
        logger.info("Preparing images and labels")
        self.images = np.asarray(list(self.crossings_images) + list(self.individual_blobs_images))
        self.images = np.expand_dims(self.images, axis = 3)
        self.labels = np.concatenate([self.crossing_labels, self.individual_blobs_labels], axis = 0)
        self.labels = self.dense_to_one_hot(np.expand_dims(self.labels, axis = 1))
        assert len(self.images) == len(self.labels)
        if self.scope == 'training':
            self.weight_positive = 1 - len(self.crossing_labels)/len(self.labels)
        else:
            self.weight_positive = 1
        np.random.seed(0)
        permutation = np.random.permutation(len(self.labels))
        self.images = self.images[permutation]
        self.labels = self.labels[permutation]

    def compute_resampling_factor(self):
        if not hasattr(self, 'resampling_factor'):
            self.resampling_factor = self.dataset_image_size / self.image_size

    def generate_crossing_images(self):
        crossing_images = []
        self.compute_resampling_factor()
        logger.debug("resampling factor crossings: %s" %str(self.resampling_factor))

        for crossing in self.crossings_sliced:
            _, _, _, crossing_image = crossing.get_image_for_identification(self.video, image_size = self.image_size)
            crossing_image = cv2.resize(crossing_image, None,
                                        fx = self.resampling_factor,
                                        fy = self.resampling_factor,
                                        interpolation = cv2.INTER_CUBIC)
            crossing_image = ((crossing_image - np.mean(crossing_image))/np.std(crossing_image)).astype('float32')
            crossing_images.append(crossing_image)

        return crossing_images

    def generate_individual_blobs_images(self):
        individual_blobs_images = []
        self.compute_resampling_factor()
        logger.debug("resampling factor individual: %s" %str(self.resampling_factor))

        for individual_blobs in self.individual_blobs_sliced:
            _, _, _, individual_blobs_image = individual_blobs.get_image_for_identification(self.video, image_size = self.image_size)
            individual_blobs_image = cv2.resize(individual_blobs_image, None,
                                                fx = self.resampling_factor,
                                                fy = self.resampling_factor,
                                                interpolation = cv2.INTER_CUBIC)
            individual_blobs_image = ((individual_blobs_image - np.mean(individual_blobs_image))/np.std(individual_blobs_image)).astype('float32')
            individual_blobs_images.append(individual_blobs_image)

        return individual_blobs_images

    def generate_test_images(self, interval = None):
        test_images = []
        self.compute_resampling_factor()
        logger.debug("resampling factor test: %s" %str(self.resampling_factor))
        if interval is None:
            blobs = self.test
        else:
            blobs = self.test[interval[0]:interval[1]]
        for blob in blobs:
            _, _, _, test_image = blob.get_image_for_identification(self.video, image_size = self.image_size)
            test_image = cv2.resize(test_image, None,
                                    fx = self.resampling_factor,
                                    fy = self.resampling_factor,
                                    interpolation = cv2.INTER_CUBIC)
            test_image = ((test_image - np.mean(test_image))/np.std(test_image)).astype('float32')
            test_images.append(test_image)

        return np.expand_dims(np.asarray(test_images), axis = 3)

    @staticmethod
    def dense_to_one_hot(labels, n_classes=2):
        """Convert class labels from scalars to one-hot vectors."""
        labels = np.array(labels)
        n_labels = labels.shape[0]
        index_offset = np.arange(n_labels) * n_classes
        labels_one_hot = np.zeros((n_labels, n_classes), dtype=np.int16)
        indices = (index_offset + labels.ravel()).astype('int')
        labels_one_hot.flat[indices] = 1
        return labels_one_hot

    @staticmethod
    def slice(data, sampling_ratio_start, sampling_ratio_end):
        num_examples = len(data)
        start = int(num_examples * sampling_ratio_start)
        end = int(num_examples * sampling_ratio_end)
        return data[start : end]



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import sys
from idtrackerai.utils.cnn_utils import *
from confapp import conf

''' original model'''
def cnn_model_0(images, classes, width, height, channels):
    '''
    Gives predictions for a given set of images
    '''
    tf.summary.image('rawImages', images, max_outputs=10)
    # conv1
    filter_size1 = 5
    n_filter1 = 16
    stride1 = [1,1,1,1]
    pad1 = 'SAME'
    conv1, w1, h1, W1 = buildConv2D('conv1', width, height, 1, images, filter_size1, n_filter1, stride1, pad1)
    # relu
    relu1 = reLU('relu1', conv1)
    # maxpool2d
    stride2 = [1,2,2,1]
    pool2 = 2
    pad2 = 'SAME'
    max_pool2, w2, h2 = maxpool2d('maxpool1',w1,h1, relu1, pool2,stride2,pad2)
    d2 = n_filter1
    # conv2
    filter_size3 = 5
    n_filter3 = 64
    stride3 = [1,1,1,1]
    pad3 = 'SAME'
    conv3, w3, h3, W3 = buildConv2D('conv2', w2, h2, d2, max_pool2, filter_size3, n_filter3, stride3, pad3)
    # relu
    relu2 = reLU('relu2', conv3)
    # maxpool2d
    stride4 = [1,2,2,1]
    pool4 = 2
    pad4 = 'SAME'
    max_pool4, w4, h4 = maxpool2d('maxpool2',w3,h3, relu2, pool4,stride4,pad4)
    d4 = n_filter3
    # conv4
    filter_size5 = 5
    n_filter5 = 100
    stride5 = [1,1,1,1]
    pad5 = 'SAME'
    conv5, w5, h5, W5  = buildConv2D('conv3', w4, h4, d4, max_pool4, filter_size5, n_filter5, stride5, pad5)
    d5 = n_filter5
    # relu
    relu3 = reLU('relu3', conv5)
    # linearize weights for fully-connected layer
    resolutionS = w5 * h5
    conv5_flat = tf.reshape(relu3, [-1, resolutionS*d5], name = 'conv5_reshape')
    # fully-connected 1
    n_fc = 100
    fc_drop, WFC = buildFc('fully-connected1', conv5_flat, w5, h5, d5, n_fc, conf.KEEP_PROB)
    relu = reLU('relu4', fc_drop)
    y_logits, WSoft = buildSoftMax('fully_connected_pre_softmax', relu, n_fc, classes)

    # return y_logits
    return y_logits, relu

''' model with 1 convolution '''
def cnn_model_1(images, classes, width, height, channels):
    '''
    Gives predictions for a given set of images
    '''

    tf.summary.image('rawImages', images, max_outputs=10)
    # conv1
    filter_size1 = 5
    n_filter1 = 16
    stride1 = [1,1,1,1]
    pad1 = 'SAME'
    conv1, w1, h1, W1 = buildConv2D('conv1', width, height, 1, images, filter_size1, n_filter1, stride1, pad1)
    # relu
    relu1 = reLU('relu1', conv1)
    # linearize weights for fully-connected layer
    resolutionS = w1 * h1
    conv1_flat = tf.reshape(relu1, [-1, resolutionS*n_filter1], name = 'conv5_reshape')
    # fully-connected 1
    n_fc = 100
    fc_drop, WFC = buildFc('fully-connected1', conv1_flat, w1, h1, n_filter1, n_fc, conf.KEEP_PROB)
    relu = reLU('relu4', fc_drop)
    y_logits, WSoft = buildSoftMax('fully_connected_pre_softmax', relu, n_fc, classes)

    return y_logits, relu
    # return y_logits, relu, (W1, W3, W5, WFC, WSoft)

''' model with 2 convolutions '''
def cnn_model_2(images, classes, width, height, channels):
    '''
    Gives predictions for a given set of images
    '''

    tf.summary.image('rawImages', images, max_outputs=10)
    # conv1
    filter_size1 = 5
    n_filter1 = 16
    stride1 = [1,1,1,1]
    pad1 = 'SAME'
    conv1, w1, h1, W1 = buildConv2D('conv1', width, height, 1, images, filter_size1, n_filter1, stride1, pad1)
    # relu
    relu1 = reLU('relu1', conv1)
    # maxpool2d
    stride2 = [1,2,2,1]
    pool2 = 2
    pad2 = 'SAME'
    max_pool2, w2, h2 = maxpool2d('maxpool1',w1,h1, relu1, pool2,stride2,pad2)
    d2 = n_filter1
    # conv2
    filter_size3 = 5
    n_filter3 = 64
    stride3 = [1,1,1,1]
    pad3 = 'SAME'
    conv3, w3, h3, W3 = buildConv2D('conv2', w2, h2, d2, max_pool2, filter_size3, n_filter3, stride3, pad3)
    # relu
    relu2 = reLU('relu2', conv3)
    # linearize weights for fully-connected layer
    resolutionS = w3 * h3
    conv3_flat = tf.reshape(relu2, [-1, resolutionS*n_filter3], name = 'conv5_reshape')
    # fully-connected 1
    n_fc = 100
    fc_drop, WFC = buildFc('fully-connected1', conv3_flat, w3, h3, n_filter3, n_fc, conf.KEEP_PROB)
    relu = reLU('relu4', fc_drop)
    y_logits, WSoft = buildSoftMax('fully_connected_pre_softmax', relu, n_fc, classes)

    return y_logits, relu
    # return y_logits, relu, (W1, W3, W5, WFC, WSoft)

''' 4 conv'''
def cnn_model_3(images, classes, width, height, channels):
    '''
    Gives predictions for a given set of images
    '''

    tf.summary.image('rawImages', images, max_outputs=10)
    # conv1
    filter_size1 = 5
    n_filter1 = 16
    stride1 = [1,1,1,1]
    pad1 = 'SAME'
    conv1, w1, h1, W1 = buildConv2D('conv1', width, height, 1, images, filter_size1, n_filter1, stride1, pad1)
    # relu
    relu1 = reLU('relu1', conv1)
    # maxpool2d
    stride2 = [1,2,2,1]
    pool2 = 2
    pad2 = 'SAME'
    max_pool2, w2, h2 = maxpool2d('maxpool1',w1,h1, relu1, pool2,stride2,pad2)
    d2 = n_filter1
    # conv2
    filter_size3 = 5
    n_filter3 = 64
    stride3 = [1,1,1,1]
    pad3 = 'SAME'
    conv3, w3, h3, W3 = buildConv2D('conv2', w2, h2, d2, max_pool2, filter_size3, n_filter3, stride3, pad3)
    # relu
    relu2 = reLU('relu2', conv3)
    # maxpool2d
    stride4 = [1,2,2,1]
    pool4 = 2
    pad4 = 'SAME'
    max_pool4, w4, h4 = maxpool2d('maxpool2',w3,h3, relu2, pool4,stride4,pad4)
    d4 = n_filter3
    # conv4
    filter_size5 = 5
    n_filter5 = 100
    stride5 = [1,1,1,1]
    pad5 = 'SAME'
    conv5, w5, h5, W5  = buildConv2D('conv3', w4, h4, d4, max_pool4, filter_size5, n_filter5, stride5, pad5)
    d5 = n_filter5
    # relu
    relu3 = reLU('relu3', conv5)
    # maxpool2d
    stride6 = [1,2,2,1]
    pool6 = 2
    pad6 = 'SAME'
    max_pool6, w6, h6 = maxpool2d('maxpool3',w5,h5, relu3, pool6,stride6,pad6)
    d6 = n_filter5
    # conv4
    filter_size7 = 5
    n_filter7 = 100
    stride7 = [1,1,1,1]
    pad7 = 'SAME'
    conv7, w7, h7, W7  = buildConv2D('conv4', w6, h6, d6, max_pool6, filter_size7, n_filter7, stride7, pad7)
    d7 = n_filter7
    # relu
    relu8 = reLU('relu4', conv7)
    # linearize weights for fully-connected layer
    resolutionS = w7 * h7
    conv7_flat = tf.reshape(relu8, [-1, resolutionS*d7], name = 'conv7_reshape')
    # fully-connected 1
    n_fc = 100
    fc_drop, WFC = buildFc('fully-connected1', conv7_flat, w7, h7, d7, n_fc, conf.KEEP_PROB)
    relu = reLU('relu4', fc_drop)
    y_logits, WSoft = buildSoftMax('fully_connected_pre_softmax', relu, n_fc, classes)

    return y_logits, relu
    # return y_logits, relu, (W1, W3, W5, WFC, WSoft)

''' inverted model '''
def cnn_model_4(images, classes, width, height, channels):
    '''
    Gives predictions for a given set of images
    '''

    tf.summary.image('rawImages', images, max_outputs=10)
    # conv1
    filter_size1 = 5
    n_filter1 = 100
    stride1 = [1,1,1,1]
    pad1 = 'SAME'
    conv1, w1, h1, W1 = buildConv2D('conv1', width, height, 1, images, filter_size1, n_filter1, stride1, pad1)
    # relu
    relu1 = reLU('relu1', conv1)
    # maxpool2d
    stride2 = [1,2,2,1]
    pool2 = 2
    pad2 = 'SAME'
    max_pool2, w2, h2 = maxpool2d('maxpool1',w1,h1, relu1, pool2,stride2,pad2)
    d2 = n_filter1
    # conv2
    filter_size3 = 5
    n_filter3 = 64
    stride3 = [1,1,1,1]
    pad3 = 'SAME'
    conv3, w3, h3, W3 = buildConv2D('conv2', w2, h2, d2, max_pool2, filter_size3, n_filter3, stride3, pad3)
    # relu
    relu2 = reLU('relu2', conv3)
    # maxpool2d
    stride4 = [1,2,2,1]
    pool4 = 2
    pad4 = 'SAME'
    max_pool4, w4, h4 = maxpool2d('maxpool2',w3,h3, relu2, pool4,stride4,pad4)
    d4 = n_filter3
    # conv4
    filter_size5 = 5
    n_filter5 = 16
    stride5 = [1,1,1,1]
    pad5 = 'SAME'
    conv5, w5, h5, W5  = buildConv2D('conv3', w4, h4, d4, max_pool4, filter_size5, n_filter5, stride5, pad5)
    d5 = n_filter5
    # relu
    relu3 = reLU('relu3', conv5)
    # linearize weights for fully-connected layer
    resolutionS = w5 * h5
    conv5_flat = tf.reshape(relu3, [-1, resolutionS*d5], name = 'conv5_reshape')
    # fully-connected 1
    n_fc = 100
    fc_drop, WFC = buildFc('fully-connected1', conv5_flat, w5, h5, d5, n_fc, conf.KEEP_PROB)
    relu = reLU('relu4', fc_drop)
    y_logits, WSoft = buildSoftMax('fully_connected_pre_softmax', relu, n_fc, classes)

    return y_logits, relu
    # return y_logits, relu, (W1, W3, W5, WFC, WSoft)

''' no nonlinearlities '''
def cnn_model_5(images, classes, width, height, channels):
    '''
    Gives predictions for a given set of images
    '''

    tf.summary.image('rawImages', images, max_outputs=10)
    # conv1
    filter_size1 = 5
    n_filter1 = 16
    stride1 = [1,1,1,1]
    pad1 = 'SAME'
    conv1, w1, h1, W1 = buildConv2D('conv1', width, height, 1, images, filter_size1, n_filter1, stride1, pad1)
    # maxpool2d
    stride2 = [1,2,2,1]
    pool2 = 2
    pad2 = 'SAME'
    max_pool2, w2, h2 = maxpool2d('maxpool1',w1,h1, conv1, pool2,stride2,pad2)
    d2 = n_filter1
    # conv2
    filter_size3 = 5
    n_filter3 = 64
    stride3 = [1,1,1,1]
    pad3 = 'SAME'
    conv3, w3, h3, W3 = buildConv2D('conv2', w2, h2, d2, max_pool2, filter_size3, n_filter3, stride3, pad3)
    # maxpool2d
    stride4 = [1,2,2,1]
    pool4 = 2
    pad4 = 'SAME'
    max_pool4, w4, h4 = maxpool2d('maxpool2',w3,h3, conv3, pool4,stride4,pad4)
    d4 = n_filter3
    # conv4
    filter_size5 = 5
    n_filter5 = 100
    stride5 = [1,1,1,1]
    pad5 = 'SAME'
    conv5, w5, h5, W5  = buildConv2D('conv3', w4, h4, d4, max_pool4, filter_size5, n_filter5, stride5, pad5)
    d5 = n_filter5
    # linearize weights for fully-connected layer
    resolutionS = w5 * h5
    conv5_flat = tf.reshape(conv5, [-1, resolutionS*d5], name = 'conv5_reshape')
    # fully-connected 1
    n_fc = 100
    fc_drop, WFC = buildFc('fully-connected1', conv5_flat, w5, h5, d5, n_fc, conf.KEEP_PROB)
    y_logits, WSoft = buildSoftMax('fully_connected_pre_softmax', fc_drop, n_fc, classes)

    return y_logits, relu
    # return y_logits, relu, (W1, W3, W5, WFC, WSoft)

''' smaller fully connected '''
def cnn_model_6(images, classes, width, height, channels):
    '''
    Gives predictions for a given set of images
    '''
    tf.summary.image('rawImages', images, max_outputs=10)
    # conv1
    filter_size1 = 5
    n_filter1 = 16
    stride1 = [1,1,1,1]
    pad1 = 'SAME'
    conv1, w1, h1, W1 = buildConv2D('conv1', width, height, 1, images, filter_size1, n_filter1, stride1, pad1)
    # relu
    relu1 = reLU('relu1', conv1)
    # maxpool2d
    stride2 = [1,2,2,1]
    pool2 = 2
    pad2 = 'SAME'
    max_pool2, w2, h2 = maxpool2d('maxpool1',w1,h1, relu1, pool2,stride2,pad2)
    d2 = n_filter1
    # conv2
    filter_size3 = 5
    n_filter3 = 64
    stride3 = [1,1,1,1]
    pad3 = 'SAME'
    conv3, w3, h3, W3 = buildConv2D('conv2', w2, h2, d2, max_pool2, filter_size3, n_filter3, stride3, pad3)
    # relu
    relu2 = reLU('relu2', conv3)
    # maxpool2d
    stride4 = [1,2,2,1]
    pool4 = 2
    pad4 = 'SAME'
    max_pool4, w4, h4 = maxpool2d('maxpool2',w3,h3, relu2, pool4,stride4,pad4)
    d4 = n_filter3
    # conv4
    filter_size5 = 5
    n_filter5 = 100
    stride5 = [1,1,1,1]
    pad5 = 'SAME'
    conv5, w5, h5, W5  = buildConv2D('conv3', w4, h4, d4, max_pool4, filter_size5, n_filter5, stride5, pad5)
    d5 = n_filter5
    # relu
    relu3 = reLU('relu3', conv5)
    # linearize weights for fully-connected layer
    resolutionS = w5 * h5
    conv5_flat = tf.reshape(relu3, [-1, resolutionS*d5], name = 'conv5_reshape')
    # fully-connected 1
    n_fc = 50
    fc_drop, WFC = buildFc('fully-connected1', conv5_flat, w5, h5, d5, n_fc, conf.KEEP_PROB)
    relu = reLU('relu4', fc_drop)
    y_logits, WSoft = buildSoftMax('fully_connected_pre_softmax', relu, n_fc, classes)

    return y_logits, relu
    # return y_logits, relu, (W1, W3, W5, WFC, WSoft)

''' bigger fully connected '''
def cnn_model_7(images, classes, width, height, channels):
    '''
    Gives predictions for a given set of images
    '''
    tf.summary.image('rawImages', images, max_outputs=10)
    # conv1
    filter_size1 = 5
    n_filter1 = 16
    stride1 = [1,1,1,1]
    pad1 = 'SAME'
    conv1, w1, h1, W1 = buildConv2D('conv1', width, height, 1, images, filter_size1, n_filter1, stride1, pad1)
    # relu
    relu1 = reLU('relu1', conv1)
    # maxpool2d
    stride2 = [1,2,2,1]
    pool2 = 2
    pad2 = 'SAME'
    max_pool2, w2, h2 = maxpool2d('maxpool1',w1,h1, relu1, pool2,stride2,pad2)
    d2 = n_filter1
    # conv2
    filter_size3 = 5
    n_filter3 = 64
    stride3 = [1,1,1,1]
    pad3 = 'SAME'
    conv3, w3, h3, W3 = buildConv2D('conv2', w2, h2, d2, max_pool2, filter_size3, n_filter3, stride3, pad3)
    # relu
    relu2 = reLU('relu2', conv3)
    # maxpool2d
    stride4 = [1,2,2,1]
    pool4 = 2
    pad4 = 'SAME'
    max_pool4, w4, h4 = maxpool2d('maxpool2',w3,h3, relu2, pool4,stride4,pad4)
    d4 = n_filter3
    # conv4
    filter_size5 = 5
    n_filter5 = 100
    stride5 = [1,1,1,1]
    pad5 = 'SAME'
    conv5, w5, h5, W5  = buildConv2D('conv3', w4, h4, d4, max_pool4, filter_size5, n_filter5, stride5, pad5)
    d5 = n_filter5
    # relu
    relu3 = reLU('relu3', conv5)
    # linearize weights for fully-connected layer
    resolutionS = w5 * h5
    conv5_flat = tf.reshape(relu3, [-1, resolutionS*d5], name = 'conv5_reshape')
    # fully-connected 1
    n_fc = 200
    fc_drop, WFC = buildFc('fully-connected1', conv5_flat, w5, h5, d5, n_fc, conf.KEEP_PROB)
    relu = reLU('relu4', fc_drop)
    y_logits, WSoft = buildSoftMax('fully_connected_pre_softmax', relu, n_fc, classes)
    return y_logits, relu
    # return y_logits, relu, (W1, W3, W5, WFC, WSoft)

''' two fully connected '''
def cnn_model_8(images, classes, width, height, channels):
    '''
    Gives predictions for a given set of images
    '''
    tf.summary.image('rawImages', images, max_outputs=10)
    # conv1
    filter_size1 = 5
    n_filter1 = 16
    stride1 = [1,1,1,1]
    pad1 = 'SAME'
    conv1, w1, h1, W1 = buildConv2D('conv1', width, height, 1, images, filter_size1, n_filter1, stride1, pad1)
    # relu
    relu1 = reLU('relu1', conv1)
    # maxpool2d
    stride2 = [1,2,2,1]
    pool2 = 2
    pad2 = 'SAME'
    max_pool2, w2, h2 = maxpool2d('maxpool1',w1,h1, relu1, pool2,stride2,pad2)
    d2 = n_filter1
    # conv2
    filter_size3 = 5
    n_filter3 = 64
    stride3 = [1,1,1,1]
    pad3 = 'SAME'
    conv3, w3, h3, W3 = buildConv2D('conv2', w2, h2, d2, max_pool2, filter_size3, n_filter3, stride3, pad3)
    # relu
    relu2 = reLU('relu2', conv3)
    # maxpool2d
    stride4 = [1,2,2,1]
    pool4 = 2
    pad4 = 'SAME'
    max_pool4, w4, h4 = maxpool2d('maxpool2',w3,h3, relu2, pool4,stride4,pad4)
    d4 = n_filter3
    # conv4
    filter_size5 = 5
    n_filter5 = 100
    stride5 = [1,1,1,1]
    pad5 = 'SAME'
    conv5, w5, h5, W5  = buildConv2D('conv3', w4, h4, d4, max_pool4, filter_size5, n_filter5, stride5, pad5)
    d5 = n_filter5
    # relu
    relu3 = reLU('relu3', conv5)
    # linearize weights for fully-connected layer
    resolutionS = w5 * h5
    conv5_flat = tf.reshape(relu3, [-1, resolutionS*d5], name = 'conv5_reshape')
    # fully-connected 1
    n_fc_1 = 100
    fc_drop_1, WFC = buildFc('fully-connected1', conv5_flat, w5, h5, d5, n_fc_1, conf.KEEP_PROB)
    relu = reLU('relu4', fc_drop_1)
    n_fc_2 = 100
    fc_drop_2, WFC_2 = buildSoftMax('fully-connected2', relu, n_fc_1, n_fc_2)
    relu = reLU('relu5', fc_drop_2)
    y_logits, WSoft = buildSoftMax('fully_connected_pre_softmax', relu, n_fc_2, classes)

    return y_logits, relu
    # return y_logits, relu, (W1, W3, W5, WFC, WSoft)

''' two fully smaller connected '''
def cnn_model_9(images, classes, width, height, channels):
    '''
    Gives predictions for a given set of images
    '''
    tf.summary.image('rawImages', images, max_outputs=10)
    # conv1
    filter_size1 = 5
    n_filter1 = 16
    stride1 = [1,1,1,1]
    pad1 = 'SAME'
    conv1, w1, h1, W1 = buildConv2D('conv1', width, height, 1, images, filter_size1, n_filter1, stride1, pad1)
    # relu
    relu1 = reLU('relu1', conv1)
    # maxpool2d
    stride2 = [1,2,2,1]
    pool2 = 2
    pad2 = 'SAME'
    max_pool2, w2, h2 = maxpool2d('maxpool1',w1,h1, relu1, pool2,stride2,pad2)
    d2 = n_filter1
    # conv2
    filter_size3 = 5
    n_filter3 = 64
    stride3 = [1,1,1,1]
    pad3 = 'SAME'
    conv3, w3, h3, W3 = buildConv2D('conv2', w2, h2, d2, max_pool2, filter_size3, n_filter3, stride3, pad3)
    # relu
    relu2 = reLU('relu2', conv3)
    # maxpool2d
    stride4 = [1,2,2,1]
    pool4 = 2
    pad4 = 'SAME'
    max_pool4, w4, h4 = maxpool2d('maxpool2',w3,h3, relu2, pool4,stride4,pad4)
    d4 = n_filter3
    # conv4
    filter_size5 = 5
    n_filter5 = 100
    stride5 = [1,1,1,1]
    pad5 = 'SAME'
    conv5, w5, h5, W5  = buildConv2D('conv3', w4, h4, d4, max_pool4, filter_size5, n_filter5, stride5, pad5)
    d5 = n_filter5
    # relu
    relu3 = reLU('relu3', conv5)
    # linearize weights for fully-connected layer
    resolutionS = w5 * h5
    conv5_flat = tf.reshape(relu3, [-1, resolutionS*d5], name = 'conv5_reshape')
    # fully-connected 1
    n_fc_1 = 100
    fc_drop_1, WFC = buildFc('fully-connected1', conv5_flat, w5, h5, d5, n_fc_1, conf.KEEP_PROB)
    relu = reLU('relu4', fc_drop_1)
    n_fc_2 = 50
    fc_drop_2, WFC_2 = buildSoftMax('fully-connected2', relu, n_fc_1, n_fc_2)
    relu = reLU('relu5', fc_drop_2)
    y_logits, WSoft = buildSoftMax('fully_connected_pre_softmax', relu, n_fc_2, classes)

    return y_logits, relu
    # return y_logits, relu, (W1, W3, W5, WFC, WSoft)

''' two fully bigger connected '''
def cnn_model_10(images, classes, width, height, channels):
    '''
    Gives predictions for a given set of images
    '''
    tf.summary.image('rawImages', images, max_outputs=10)
    # conv1
    filter_size1 = 5
    n_filter1 = 16
    stride1 = [1,1,1,1]
    pad1 = 'SAME'
    conv1, w1, h1, W1 = buildConv2D('conv1', width, height, 1, images, filter_size1, n_filter1, stride1, pad1)
    # relu
    relu1 = reLU('relu1', conv1)
    # maxpool2d
    stride2 = [1,2,2,1]
    pool2 = 2
    pad2 = 'SAME'
    max_pool2, w2, h2 = maxpool2d('maxpool1',w1,h1, relu1, pool2,stride2,pad2)
    d2 = n_filter1
    # conv2
    filter_size3 = 5
    n_filter3 = 64
    stride3 = [1,1,1,1]
    pad3 = 'SAME'
    conv3, w3, h3, W3 = buildConv2D('conv2', w2, h2, d2, max_pool2, filter_size3, n_filter3, stride3, pad3)
    # relu
    relu2 = reLU('relu2', conv3)
    # maxpool2d
    stride4 = [1,2,2,1]
    pool4 = 2
    pad4 = 'SAME'
    max_pool4, w4, h4 = maxpool2d('maxpool2',w3,h3, relu2, pool4,stride4,pad4)
    d4 = n_filter3
    # conv4
    filter_size5 = 5
    n_filter5 = 100
    stride5 = [1,1,1,1]
    pad5 = 'SAME'
    conv5, w5, h5, W5  = buildConv2D('conv3', w4, h4, d4, max_pool4, filter_size5, n_filter5, stride5, pad5)
    d5 = n_filter5
    # relu
    relu3 = reLU('relu3', conv5)
    # linearize weights for fully-connected layer
    resolutionS = w5 * h5
    conv5_flat = tf.reshape(relu3, [-1, resolutionS*d5], name = 'conv5_reshape')
    # fully-connected 1
    n_fc_1 = 100
    fc_drop_1, WFC = buildFc('fully-connected1', conv5_flat, w5, h5, d5, n_fc_1, conf.KEEP_PROB)
    relu = reLU('relu4', fc_drop_1)
    n_fc_2 = 200
    fc_drop_2, WFC_2 = buildSoftMax('fully-connected2', relu, n_fc_1, n_fc_2)
    relu = reLU('relu5', fc_drop_2)
    y_logits, WSoft = buildSoftMax('fully_connected_pre_softmax', relu, n_fc_2, classes)

    return y_logits, relu
    # return y_logits, relu, (W1, W3, W5, WFC, WSoft)

''' even smaller fully connected '''
def cnn_model_11(images, classes, width, height, channels):
    '''
    Gives predictions for a given set of images
    '''
    tf.summary.image('rawImages', images, max_outputs=10)
    # conv1
    filter_size1 = 5
    n_filter1 = 16
    stride1 = [1,1,1,1]
    pad1 = 'SAME'
    conv1, w1, h1, W1 = buildConv2D('conv1', width, height, 1, images, filter_size1, n_filter1, stride1, pad1)
    # relu
    relu1 = reLU('relu1', conv1)
    # maxpool2d
    stride2 = [1,2,2,1]
    pool2 = 2
    pad2 = 'SAME'
    max_pool2, w2, h2 = maxpool2d('maxpool1',w1,h1, relu1, pool2,stride2,pad2)
    d2 = n_filter1
    # conv2
    filter_size3 = 5
    n_filter3 = 64
    stride3 = [1,1,1,1]
    pad3 = 'SAME'
    conv3, w3, h3, W3 = buildConv2D('conv2', w2, h2, d2, max_pool2, filter_size3, n_filter3, stride3, pad3)
    # relu
    relu2 = reLU('relu2', conv3)
    # maxpool2d
    stride4 = [1,2,2,1]
    pool4 = 2
    pad4 = 'SAME'
    max_pool4, w4, h4 = maxpool2d('maxpool2',w3,h3, relu2, pool4,stride4,pad4)
    d4 = n_filter3
    # conv4
    filter_size5 = 5
    n_filter5 = 100
    stride5 = [1,1,1,1]
    pad5 = 'SAME'
    conv5, w5, h5, W5  = buildConv2D('conv3', w4, h4, d4, max_pool4, filter_size5, n_filter5, stride5, pad5)
    d5 = n_filter5
    # relu
    relu3 = reLU('relu3', conv5)
    # linearize weights for fully-connected layer
    resolutionS = w5 * h5
    conv5_flat = tf.reshape(relu3, [-1, resolutionS*d5], name = 'conv5_reshape')
    # fully-connected 1
    n_fc = 10
    fc_drop, WFC = buildFc('fully-connected1', conv5_flat, w5, h5, d5, n_fc, conf.KEEP_PROB)
    relu = reLU('relu4', fc_drop)
    y_logits, WSoft = buildSoftMax('fully_connected_pre_softmax', relu, n_fc, classes)

    return y_logits, relu
    # return y_logits, relu, (W1, W3, W5, WFC, WSoft)

def cnn_model_crossing_detector(images, classes, width, height, channels):
    '''
    Gives predictions for a given set of images
    '''
    tf.summary.image('rawImages', images, max_outputs=10)
    # conv1
    filter_size1 = 5
    n_filter1 = 16
    stride1 = [1,1,1,1]
    pad1 = 'SAME'
    conv1, w1, h1, W1 = buildConv2D('conv1', width, height, 1, images, filter_size1, n_filter1, stride1, pad1)
    # relu
    relu1 = reLU('relu1', conv1)
    # maxpool2d
    stride2 = [1,2,2,1]
    pool2 = 2
    pad2 = 'SAME'
    max_pool2, w2, h2 = maxpool2d('maxpool1',w1,h1, relu1, pool2,stride2,pad2)
    d2 = n_filter1
    # conv2
    filter_size3 = 5
    n_filter3 = 64
    stride3 = [1,1,1,1]
    pad3 = 'SAME'
    conv3, w3, h3, W3 = buildConv2D('conv2', w2, h2, d2, max_pool2, filter_size3, n_filter3, stride3, pad3)
    # relu
    relu2 = reLU('relu2', conv3)
    # linearize weights for fully-connected layer
    resolutionS = w3 * h3
    conv5_flat = tf.reshape(relu2, [-1, resolutionS*n_filter3], name = 'conv5_reshape')
    # fully-connected 1
    n_fc = 100
    fc_drop, WFC = buildFc('fully-connected1', conv5_flat, w3, h3, n_filter3, n_fc, conf.KEEP_PROB)
    relu = reLU('relu1', fc_drop)
    y_logits, WSoft = buildSoftMax('fully_connected_pre_softmax', relu, n_fc, classes)
    return y_logits



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import os
import sys

import h5py
import numpy as np
from tqdm import tqdm

from idtrackerai.fragment import Fragment
from idtrackerai.utils.py_utils import set_attributes_of_object_to_value, append_values_to_lists

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.list_of_fragments")

class ListOfFragments(object):
    """ Collects all the instances of the class :class:`~fragment.Fragment`
    generated from the blobs extracted from the video during segmentation
    (see :mod:`~segmentation`) after having assigned to each Blob instance
    a fragment identifier by using the method
    :meth:`~list_of_blobs.compute_fragment_identifier_and_blob_index`

    Attributes
    ----------

    fragments : list
        list of instances of the class :class:`~fragment.Fragment`
    number_of_fragments : int
        number of fragments computed by the method
        :meth:`~list_of_blobs.compute_fragment_identifier_and_blob_index`
    """
    def __init__(self, fragments, identification_images_file_path):
        self.fragments = fragments
        self.number_of_fragments = len(self.fragments)
        self.identification_images_file_path = identification_images_file_path

    def get_fragment_identifier_to_index_list(self):
        """Creates a mapping between the attribute :attr:`fragments` and
        their identifiers build from the :class:`~list_of_blobs.ListOfBlobs`

        Returns
        -------
        list
            Mapping from the collection of fragments to the list of fragment
            identifiers

        """
        fragments_identifiers = [fragment.identifier for fragment in self.fragments]
        fragment_identifier_to_index = np.arange(len(fragments_identifiers))
        fragments_identifiers_argsort = np.argsort(fragments_identifiers)
        return fragment_identifier_to_index[fragments_identifiers_argsort]

    def reset(self, roll_back_to = None):
        """Resets all the fragment by using the method
        :meth:`~fragment.Fragment.roll_back_to`
        """
        logger.warning("Reseting list_of_fragments")
        for fragment in self.fragments:
            fragment.reset(roll_back_to)
        logger.warning("Done")

    def get_images_from_fragments_to_assign(self):
        """Take all the fragments that have not been used to train the idCNN
        and that are associated with an individual, and concatenates their
        images in order to feed them to the idCNN and get predictions

        Returns
        -------
        ndarray
            [number_of_images, height, width, number_of_channels]
        """
        images_lists = [fragment.images for fragment in self.fragments
                        if not fragment.used_for_training and fragment.is_an_individual]
        images = [image for images in images_lists for image in images]
        return np.asarray(load_identification_images(self.identification_images_file_path, images))


    def compute_number_of_unique_images_used_for_pretraining(self):
        """Returns the number of images used for pretraining
        (without repetitions)

        Returns
        -------
        int
            Number of images used in pretraining
        """
        return sum([fragment.number_of_images for fragment in self.fragments if fragment.used_for_pretraining])

    def compute_number_of_unique_images_used_for_training(self):
        """Returns the number of images used for training
        (without repetitions)

        Returns
        -------
        int
            Number of images used in training
        """
        return sum([fragment.number_of_images for fragment in self.fragments if fragment.used_for_training])

    def compute_total_number_of_images_in_global_fragments(self):
        """Sets the number of images available in global fragments (without repetitions)
        """
        self.number_of_images_in_global_fragments = sum([fragment.number_of_images
                                                         for fragment in self.fragments
                                                         if fragment.identifier in self.accumulable_individual_fragments])
        return self.number_of_images_in_global_fragments

    def compute_ratio_of_images_used_for_pretraining(self):
        """Returns the ratio of images used for pretraining over the number of
        available images

        Returns
        -------
        float
            Ratio of images used for pretraining
        """
        return self.compute_number_of_unique_images_used_for_pretraining() / self.number_of_images_in_global_fragments

    def compute_ratio_of_images_used_for_training(self):
        """Returns the ratio of images used for training over the number of
        available images

        Returns
        -------
        float
            Ratio of images used for training
        """
        return self.compute_number_of_unique_images_used_for_training() / self.number_of_images_in_global_fragments

    def compute_P2_vectors(self):
        """Computes the P2_vector associated to every individual fragment. See
        :meth:`~fragment.Fragment.compute_P2_vector`
        """
        [fragment.compute_P2_vector() for fragment in self.fragments if fragment.is_an_individual]

    def get_number_of_unidentified_individual_fragments(self):
        """Returns the number of individual fragments that have not been
        identified during the fingerprint protocols cascade

        Returns
        -------
        int
            number of non-identified individual fragments
        """
        return len([fragment for fragment in self.fragments if fragment.is_an_individual and not fragment.used_for_training])

    def get_next_fragment_to_identify(self):
        """Returns the next fragment to be identified after the fingerprint
        protocols cascade by sorting according to the certainty computed with
        P2. See :attr:`~fragment.Fragment.certainty_P2`

        Returns
        -------
        <Fragment object>
            an instance of the class :class:`~fragment.Fragment`

        """
        fragments = [fragment for fragment in self.fragments if fragment.is_an_individual and fragment.assigned_identities[0] is None]
        fragments.sort(key=lambda x: x.certainty_P2, reverse=True)
        return fragments[0]

    def update_identification_images_dataset(self):
        with h5py.File(self.identification_images_file_path, 'a') as f:
            f.create_dataset("identities", (f['identification_images'].shape[0], 1),
                             fillvalue=np.nan)
            for fragment in self.fragments:
                if fragment.used_for_training:
                    for image in fragment.images:
                        f['identities'][image] = fragment.identity

    def get_ordered_list_of_fragments(self, scope = None, first_frame_first_global_fragment = None):
        """Sorts the fragments starting from the frame number
        `first_frame_first_global_fragment`. According to `scope` the sorting
        is done either "to the future" of "to the past" with respect to
        `first_frame_first_global_fragment`

        Parameters
        ----------
        scope : str
            either "to_the_past" or "to_the_future"
        first_frame_first_global_fragment : int
            frame number corresponding to the first frame in which all the
            individual fragments coexist in the first global fragment using
            in an iteration of the fingerprint protocol cascade

        Returns
        -------
        list
            list of sorted fragments

        """
        if scope == 'to_the_past':
            fragments_subset = [fragment for fragment in self.fragments if fragment.start_end[1] <= first_frame_first_global_fragment]
            fragments_subset.sort(key=lambda x: x.start_end[1], reverse=True)
        elif scope == 'to_the_future':
            fragments_subset = [fragment for fragment in self.fragments if fragment.start_end[0] >= first_frame_first_global_fragment]
            fragments_subset.sort(key=lambda x: x.start_end[0], reverse=False)
        return fragments_subset

    def save(self, fragments_path):
        """saves an instance of ListOfFragments in the path specified by
        `fragments_path`
        """
        logger.info("saving list of fragments at %s" %fragments_path)
        for fragment in self.fragments:
            fragment.coexisting_individual_fragments = None
        np.save(fragments_path,self)
        for fragment in self.fragments:
            fragment.get_coexisting_individual_fragments_indices(self.fragments)

    @classmethod
    def load(cls, path_to_load):
        """Loads a previously saved (see :meth:`load`) from the path
        `path_to_load`
        """
        logger.info("loading list of fragments from %s" %path_to_load)
        list_of_fragments = np.load(path_to_load, allow_pickle=True).item()
        for fragment in list_of_fragments.fragments:
            fragment.get_coexisting_individual_fragments_indices(list_of_fragments.fragments)
        return list_of_fragments

    def create_light_list(self, attributes = None):
        """Creates a light version of an instance of ListOfFragments by storing
        only the attributes listed in `attributes` in a list of dictionaries

        Parameters
        ----------
        attributes : list
            list of attributes to be stored

        Returns
        -------
        list
            list of dictionaries organised per fragment with keys the
            attributes listed in `attributes`
        """
        if attributes == None:
            attributes_to_discard = ['images', 'coexisting_individual_fragments']
        return [{attribute: getattr(fragment, attribute) for attribute in fragment.__dict__.keys() if attribute not in attributes_to_discard}
                    for fragment in self.fragments]

    def save_light_list(self, accumulation_folder):
        """Saves a list of dictionaries created with the method
        :meth:`create_light_list` in the folder `accumulation_folder`
        """
        np.save(os.path.join(accumulation_folder, 'light_list_of_fragments.npy'), self.create_light_list())

    def load_light_list(self, accumulation_folder):
        """Loads a list of dictionaries created with the method
        :meth:`create_light_list` and saved with :meth:`save_light_list` from
        the folder `accumulation_folder`
        """
        list_of_dictionaries = np.load(os.path.join(accumulation_folder, 'light_list_of_fragments.npy'), allow_pickle=True)
        self.update_fragments_dictionary(list_of_dictionaries)

    def update_fragments_dictionary(self, list_of_dictionaries):
        """Update fragment objects (see :class:`~fragment.Fragment`) by
        considering a list of dictionaries
        """
        assert len(list_of_dictionaries) == len(self.fragments)
        [fragment.__dict__.update(dictionary) for fragment, dictionary in zip(self.fragments, list_of_dictionaries)]

    def get_new_images_and_labels_for_training(self):
        """Extract images and creates labels from every individual fragment
        that has not been used to train the network during the fingerprint
        protocols cascade

        Returns
        -------
        list
            images
        list
            labels
        """
        images = []
        labels = []
        for fragment in self.fragments:
            if fragment.acceptable_for_training and not fragment.used_for_training:
                assert fragment.is_an_individual
                images.extend(fragment.images)
                labels.extend([fragment.temporary_id] * fragment.number_of_images)
        if len(images) != 0:
            return images, np.asarray(labels)
        else:
            return None, None

    def get_accumulable_individual_fragments_identifiers(self, list_of_global_fragments):
        """Gets the unique identifiers associated to individual fragments that
        can be accumulated

        Parameters
        ----------
        list_of_global_fragments : <ListOfGlobalFragments object>
            Object collecting the global fragment objects (instances of the
            class :class:`~global_fragment.GlobalFragment`) detected in the
            entire video

        """
        self.accumulable_individual_fragments = set([identifier for global_fragment in list_of_global_fragments.global_fragments
                                                                        for identifier in global_fragment.individual_fragments_identifiers])

    def get_not_accumulable_individual_fragments_identifiers(self, list_of_global_fragments):
        """Gets the unique identifiers associated to individual fragments that
        cannot be accumulated

        Parameters
        ----------
        list_of_global_fragments : <ListOfGlobalFragments object>
            Object collecting the global fragment objects (instances of the
            class :class:`~global_fragment.GlobalFragment`) detected in the
            entire video

        """
        self.not_accumulable_individual_fragments = set([identifier for global_fragment in list_of_global_fragments.non_accumulable_global_fragments
                                                        for identifier in global_fragment.individual_fragments_identifiers]) - self.accumulable_individual_fragments

    def set_fragments_as_accumulable_or_not_accumulable(self):
        """Set the attribute :attr:`~fragment.accumulable`
        """
        for fragment in self.fragments:
            if fragment.identifier in self.accumulable_individual_fragments:
                setattr(fragment, '_accumulable', True)
            elif fragment.identifier in self.not_accumulable_individual_fragments:
                setattr(fragment, '_accumulable', False)
            else:
                setattr(fragment, '_accumulable', None)

    def get_stats(self, list_of_global_fragments):
        """Collects the following statistics from both fragments and global
        fragments:

        *number_of_fragments
        *number_of_crossing_fragments
        *number_of_individual_fragments
        *number_of_individual_fragments_not_in_a_global_fragment
        *number_of_accumulable_individual_fragments
        *number_of_not_accumulable_individual_fragments
        *number_of_accumualted_individual_fragments
        *number_of_globally_accumulated_individual_fragments
        *number_of_partially_accumulated_individual_fragments
        *number_of_blobs
        *number_of_crossing_blobs
        *number_of_individual_blobs
        *number_of_individual_blobs_not_in_a_global_fragment
        *number_of_accumulable_individual_blobs
        *number_of_not_accumulable_individual_blobs
        *number_of_accumualted_individual_blobs
        *number_of_globally_accumulated_individual_blobs
        *number_of_partially_accumulated_individual_blobs

        Parameters
        ----------
        list_of_global_fragments : <ListOfGlobalFragments object>
            Object collecting the global fragment objects (instances of the
            class :class:`~global_fragment.GlobalFragment`) detected in the
            entire video

        Returns
        -------
        dict
            dictionary of statistics

        """
        # number of fragments per class
        self.number_of_crossing_fragments = sum([fragment.is_a_crossing for fragment in self.fragments])
        self.number_of_individual_fragments = sum([fragment.is_an_individual for fragment in self.fragments])
        self.number_of_individual_fragments_not_in_a_global_fragment = sum([not fragment.is_in_a_global_fragment
                                                                        and fragment.is_an_individual
                                                                        for fragment in self.fragments])
        self.number_of_accumulable_individual_fragments = len(self.accumulable_individual_fragments)
        self.number_of_not_accumulable_individual_fragments = len(self.not_accumulable_individual_fragments)
        fragments_not_accumualted = set([fragment.identifier for fragment in self.fragments if not fragment.used_for_training])
        self.number_of_not_accumulated_individual_fragments = len(self.accumulable_individual_fragments & fragments_not_accumualted)
        self.number_of_globally_accumulated_individual_fragments = sum([fragment.accumulated_globally
                                                                    and fragment.is_an_individual for fragment in self.fragments])
        self.number_of_partially_accumulated_individual_fragments = sum([fragment.accumulated_partially
                                                                    and fragment.is_an_individual for fragment in self.fragments])
        # number of blobs per class
        self.number_of_blobs = sum([fragment.number_of_images for fragment in self.fragments])
        self.number_of_crossing_blobs = sum([fragment.is_a_crossing * fragment.number_of_images for fragment in self.fragments])
        self.number_of_individual_blobs = sum([fragment.is_an_individual * fragment.number_of_images for fragment in self.fragments])
        self.number_of_individual_blobs_not_in_a_global_fragment = sum([(not fragment.is_in_a_global_fragment
                                                                        and fragment.is_an_individual) * fragment.number_of_images
                                                                        for fragment in self.fragments])
        self.number_of_accumulable_individual_blobs = sum([fragment.accumulable * fragment.number_of_images for fragment in self.fragments
                                                                if fragment.accumulable is not None])
        self.number_of_not_accumulable_individual_blobs = sum([(not fragment.accumulable) * fragment.number_of_images for fragment in self.fragments
                                                                if fragment.accumulable is not None])
        fragments_not_accumualted = self.accumulable_individual_fragments & set([fragment.identifier for fragment in self.fragments if not fragment.used_for_training])
        self.number_of_not_accumulated_individual_blobs = sum([fragment.number_of_images for fragment in self.fragments if fragment.identifier in fragments_not_accumualted])
        self.number_of_globally_accumulated_individual_blobs = sum([(fragment.accumulated_globally
                                                                    and fragment.is_an_individual) * fragment.number_of_images for fragment in self.fragments if fragment.accumulated_globally is not None])
        self.number_of_partially_accumulated_individual_blobs = sum([(fragment.accumulated_partially
                                                                    and fragment.is_an_individual) * fragment.number_of_images for fragment in self.fragments if fragment.accumulated_partially is not None])

        logger.info('number_of_fragments %i' %self.number_of_fragments)
        logger.info('number_of_crossing_fragments %i' %self.number_of_crossing_fragments)
        logger.info('number_of_individual_fragments %i ' %self.number_of_individual_fragments)
        logger.info('number_of_individual_fragments_not_in_a_global_fragment %i' %self.number_of_individual_fragments_not_in_a_global_fragment)
        logger.info('number_of_accumulable_individual_fragments %i' %self.number_of_accumulable_individual_fragments)
        logger.info('number_of_not_accumulable_individual_fragments %i' %self.number_of_not_accumulable_individual_fragments)
        logger.info('number_of_not_accumulated_individual_fragments %i' %self.number_of_not_accumulated_individual_fragments)
        logger.info('number_of_globally_accumulated_individual_fragments %i' %self.number_of_globally_accumulated_individual_fragments)
        logger.info('number_of_partially_accumulated_individual_fragments %i' %self.number_of_partially_accumulated_individual_fragments)
        attributes_to_return = ['number_of_fragments',
            'number_of_crossing_fragments', 'number_of_individual_fragments',
            'number_of_individual_fragments_not_in_a_global_fragment',
            'number_of_accumulable_individual_fragments',
            'number_of_not_accumulable_individual_fragments',
            'number_of_accumualted_individual_fragments',
            'number_of_globally_accumulated_individual_fragments',
            'number_of_partially_accumulated_individual_fragments',
            'number_of_blobs', 'number_of_crossing_blobs',
            'number_of_individual_blobs',
            'number_of_individual_blobs_not_in_a_global_fragment',
            'number_of_accumulable_individual_blobs',
            'number_of_not_accumulable_individual_blobs',
            'number_of_accumualted_individual_blobs',
            'number_of_globally_accumulated_individual_blobs',
            'number_of_partially_accumulated_individual_blobs']
        return {key: getattr(self, key) for key in self.__dict__ if key in attributes_to_return}

    # def plot_stats(self, video):
    #     """Plots the statistics obtained through :meth:`get_stats`
    #
    #     Parameters
    #     ----------
    #     video : <Video object>
    #         See :class:`~video.Video`
    #     """
    #     from matplotlib import pyplot as plt
    #     import matplotlib.lines as mlines
    #     import seaborn as sns
    #     plt.ion()
    #     fig, ax = plt.subplots(1,1)
    #     sns.set_style("ticks")
    #     colors = ['grey', 'y', 'y', 'r', 'g', 'g']
    #     hatches = ['', '', '/', '', '', '//']
    #     labels = ['crossings', 'not in GF', 'not accumulable',
    #                 'not accumulated', 'globally accumulated',
    #                 'partially accumulated']
    #     sizes = np.asarray([self.number_of_crossing_blobs,
    #                         self.number_of_individual_blobs_not_in_a_global_fragment,
    #                         self.number_of_not_accumulable_individual_blobs,
    #                         self.number_of_not_accumulated_individual_blobs,
    #                         self.number_of_globally_accumulated_individual_blobs,
    #                         self.number_of_partially_accumulated_individual_blobs]) / self.number_of_blobs * 100
    #     labels_with_percentage = [' %.2f' %size + r'$\%$ - ' + label for label, size in zip(labels, sizes)]
    #     explode = (0, 0, 0, 0, 0, 0)  # only "explode" the 2nd slice (i.e. 'Hogs')
    #
    #     patches_pie = ax.pie(sizes, colors = colors, explode=explode, autopct='%1.1f%%',
    #                 shadow=False, startangle=90, pctdistance=1.1)
    #     # patterns = ('-', '+', 'x', '\\', '*', 'o', 'O', '.')
    #     for hatch, patch in zip(hatches, patches_pie[0]):
    #         patch.set_hatch(hatch)
    #     ax.axis('equal')  # Equal aspect ratio ensures that pie is drawn as a circle.
    #     ax.legend(patches_pie, labels=labels_with_percentage, loc=3, title = 'Blob class percentage')
    #
    #     fig.savefig(os.path.join(video.preprocessing_folder,'fragments_summary_1.pdf'), transparent=True)
    #
    #     import matplotlib.patches as patches
    #     def get_fragment_identification_type(fragment):
    #         """Returns the an identifier of the process in which fragment has
    #         been identified in the algorithm:
    #         *0: The fragment is a crossing, hence identified only during
    #         post-processing
    #         *1: The (individual) fragment has been identified during the
    #         fingerprint protocols cascade
    #         *2: The (individual) fragment has been identified after the
    #         protocols cascade
    #         *3: The (individual) fragment has not been assigned
    #
    #         Parameters
    #         ----------
    #         fragment : <Fragment object>
    #             an instance of the class :class:`~fragment.Fragment`
    #
    #         Returns
    #         -------
    #         int
    #             identifier of the identification state of the fragment
    #
    #         """
    #         if fragment.is_a_crossing:
    #             return 0 #crossing
    #         elif fragment.is_an_individual and (fragment.accumulated_globally or fragment.accumulated_partially):
    #             return 1 #assigned during accumulation
    #         elif fragment.identity != 0:
    #             return 2 #assigned after accumulation
    #         else:
    #             return 3 #not assigned
    #
    #     labels = ['crossings', 'assigned during accumulation', 'assigned after accumulation', 'not assigned']
    #     colors = ['k', 'g', 'y', 'r']
    #     fig, ax = plt.subplots(1,1)
    #     sns.set_style("ticks")
    #
    #     for fragment in self.fragments:
    #         if fragment.is_an_individual:
    #             blob_index = fragment.blob_hierarchy_in_starting_frame
    #             type = get_fragment_identification_type(fragment)
    #             (start, end) = fragment.start_end
    #             ax.add_patch(
    #                 patches.Rectangle(
    #                     (start, blob_index - 0.5),   # (x,y)
    #                     end - start,  # width
    #                     1.,          # height
    #                     fill=True,
    #                     edgecolor=None,
    #                     facecolor=colors[type],
    #                     alpha = 1.
    #                 )
    #             )
    #
    #     ax.axis('tight')
    #     ax.set_xlabel('Frame number')
    #     ax.set_ylabel('Blob index')
    #     ax.set_yticks(range(0,video.number_of_animals,4))
    #     ax.set_yticklabels(range(1,video.number_of_animals+1,4))
    #     ax.set_xlim([0., video.number_of_frames])
    #     ax.set_ylim([-.5, .5 + video.number_of_animals - 1])
    #     fig.savefig(os.path.join(video.preprocessing_folder,'fragments_summary_2.pdf'), transparent=True)
    #     plt.show()


def create_list_of_fragments(blobs_in_video, number_of_animals):
    """Generate a list of instances of :class:`~fragment.Fragment` collecting
    all the fragments in the video.

    Parameters
    ----------
    blobs_in_video : list
        list of the blob objects (see class :class:`~blob.Blob`) generated
        from the blobs segmented in the video
    number_of_animals : int
        Number of animals to track

    Returns
    -------
    list
        list of instances of :class:`~fragment.Fragment`

    """
    attributes_to_set = ['_image_for_identification', '_next', '_previous']
    fragments = []
    used_fragment_identifiers = set()

    for blobs_in_frame in tqdm(blobs_in_video, desc = 'creating list of fragments'):
        for blob in blobs_in_frame:
            current_fragment_identifier = blob.fragment_identifier
            if current_fragment_identifier not in used_fragment_identifiers:
                images = [blob.identification_image_index] if blob.is_an_individual else [None]
                bounding_boxes = [blob.bounding_box_in_frame_coordinates] if blob.is_a_crossing else []
                centroids = [blob.centroid]
                areas = [blob.area]
                start = blob.frame_number
                current = blob

                while len(current.next) > 0 and current.next[0].fragment_identifier == current_fragment_identifier:
                    current = current.next[0]
                    bounding_box_in_frame_coordinates = [current.bounding_box_in_frame_coordinates] if current.is_a_crossing else []
                    images, bounding_boxes, centroids, areas = append_values_to_lists([current.identification_image_index,
                                                                bounding_box_in_frame_coordinates,
                                                                current.centroid,
                                                                current.area],
                                                                [images,
                                                                bounding_boxes,
                                                                centroids,
                                                                areas])

                end = current.frame_number
                fragment = Fragment(current_fragment_identifier,
                                    (start, end + 1), # it is not inclusive to follow Python convention
                                    blob.blob_index,
                                    images,
                                    bounding_boxes,
                                    centroids,
                                    areas,
                                    blob.is_an_individual,
                                    blob.is_a_crossing,
                                    number_of_animals,
                                    user_generated_identities = blob.user_generated_identities)
                used_fragment_identifiers.add(current_fragment_identifier)
                fragments.append(fragment)

            set_attributes_of_object_to_value(blob, attributes_to_set, value = None)
    logger.info("getting coexisting individual fragments indices")
    [fragment.get_coexisting_individual_fragments_indices(fragments) for fragment in fragments]
    return fragments


def load_identification_images(identification_images_file_path, images_indices):
    with h5py.File(identification_images_file_path, 'r') as f:
        dset = f['identification_images']
        images = [dset[image_index, ...] for image_index in images_indices]
    return images



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import os
import sys
import glob
import re

import numpy as np
from itertools import groupby
import multiprocessing
import matplotlib
from matplotlib import cm

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.py_utils")

### MKL
def set_mkl_to_single_thread():
    logger.info('Setting MKL library to use single thread')
    os.environ['MKL_NUM_THREADS'] = '1'
    os.environ['OMP_NUM_THREADS'] = '1'
    os.environ['MKL_DYNAMIC'] = 'FALSE'

def set_mkl_to_multi_thread():
    logger.info('Setting MKL library to use multiple threads')
    os.environ['MKL_NUM_THREADS'] = str(multiprocessing.cpu_count())
    os.environ['OMP_NUM_THREADS'] = str(multiprocessing.cpu_count())
    os.environ['MKL_DYNAMIC'] = 'TRUE'

### Object utils ###
def append_values_to_lists(values, list_of_lists):
    list_of_lists_updated = []

    for l, value in zip(list_of_lists, values):
        l.append(value)
        list_of_lists_updated.append(l)

    return list_of_lists_updated

def set_attributes_of_object_to_value(object_to_modify, attributes_list, value = None):
    [setattr(object_to_modify, attribute, value) for attribute in attributes_list if hasattr(object_to_modify, attribute)]

def delete_attributes_from_object(object_to_modify, list_of_attributes):
    [delattr(object_to_modify, attribute) for attribute in list_of_attributes if hasattr(object_to_modify, attribute)]

### Dict utils ###
def flatten(l):
    ''' flatten a list of lists '''
    try:
        ans = [inner for outer in l for inner in outer]
    except:
        ans = [y for x in l for y in (x if isinstance(x, tuple) else (x,))]
    return ans

def natural_sort(l):
    convert = lambda text: int(text) if text.isdigit() else text.lower()
    alphanum_key = lambda key: [ convert(c) for c in re.split('([0-9]+)', key) ]
    return sorted(l, key = alphanum_key)

def scanFolder(path):
    ### NOTE if the video selected does not finish with '_1' the scanFolder function won't select all of them. This can be improved
    paths = [path]
    video_path = os.path.basename(path)
    filename, extension = os.path.splitext(video_path)
    folder = os.path.dirname(path)
    # maybe write check on video extension supported by opencv2
    if filename[-2:] == '_1':
        paths = natural_sort(glob.glob(folder + "/" + filename[:-1] + "*" + extension))
    return paths

def get_spaced_colors_util(n, norm = False, black = True, cmap = 'jet'):
    RGB_tuples = cm.get_cmap(cmap)
    if norm:
        colors = [RGB_tuples(i / n) for i in range(n)]
    else:
        RGB_array = np.asarray([RGB_tuples(i / n) for i in range(n)])
        BRG_array = np.zeros(RGB_array.shape)
        BRG_array[:,0] = RGB_array[:,2]
        BRG_array[:,1] = RGB_array[:,1]
        BRG_array[:,2] = RGB_array[:,0]
        colors = [tuple(BRG_array[i,:] * 256) for i in range(n)]
    if black:
        black = (0., 0., 0.)
        colors.insert(0, black)
    return colors

def check_and_change_video_path(video_object, old_video):
    current_video_folder = os.path.split(video_object.video_path)[0]
    old_video_folder = os.path.split(old_video.video_path)[0]
    old_video_session_name = old_video.session_folder
    print("in check and change video path")
    print("current_video_folder: ", current_video_folder)
    print("old_video_folder: ", old_video_folder)
    print("old_video_session_name: ", old_video_session_name)
    if current_video_folder != old_video_folder:
        print("updating attributes")
        attributes_to_modify = {key: getattr(old_video, key) for key in old_video.__dict__
        if isinstance(getattr(old_video, key), basestring)
        and old_video_folder in getattr(old_video, key) }

        for key in attributes_to_modify:
            new_value = attributes_to_modify[key].replace(old_video_folder, current_video_folder)
            setattr(old_video, key, new_value)

        if old_video.paths_to_video_segments is not None and len(old_video.paths_to_video_segments) != 0:
            logger.info("Updating paths_to_video_segments")
            new_paths_to_video_segments = []
            for path in old_video.paths_to_video_segments:
                new_path = os.path.join(old_video.video_folder, os.path.split(path)[1])
                new_paths_to_video_segments.append(new_path)
            old_video._paths_to_video_segments = new_paths_to_video_segments

        ### update checkpoint files
        current_video_session_name = old_video.session_folder
        folders_to_check = ['_crossings_detector_folder',
                            '_pretraining_folder',
                            '_accumulation_folder']
        for folder in folders_to_check:
            if hasattr(old_video, folder) and getattr(old_video, folder) is not None:
                if folder == 'crossings_detector_folder':
                    checkpoint_path = os.path.join(old_video.crossings_detector_folder, 'checkpoint')
                    if os.path.isfile(checkpoint_path):
                        old_video.update_tensorflow_checkpoints_file(checkpoint_path, old_video_session_name, current_video_session_name)
                    else:
                        logger.warn('No checkpoint found in %s ' %folder)
                else:
                    for sub_folder in ['conv', 'softmax']:
                        checkpoint_path = os.path.join(getattr(old_video,folder), sub_folder, 'checkpoint')
                        if os.path.isfile(checkpoint_path):
                            old_video.update_tensorflow_checkpoints_file(checkpoint_path, old_video_session_name, current_video_session_name)
                        else:
                            logger.warn('No checkpoint found in %s ' %os.path.join(getattr(old_video, folder), sub_folder))
    return old_video

def set_load_previous_dict(old_video, processes, existentFile):
    print("in set load previous dict")
    attributes = [ 'has_been_preprocessed',
                    'first_accumulation_finished',
                    'has_been_pretrained', 'second_accumulation_finished',
                    'has_been_assigned',
                    'has_crossings_solved', 'has_trajectories',
                    'has_trajectories_wo_gaps']

    # gui_processes = ['preprocessing','protocols1_and_2', 'protocol3_pretraining',
    #                 'protocol3_accumulation', 'residual_identification',
    #                 'post_processing']
    processes_to_attributes = { 'preprocessing' : ['has_been_preprocessed'],
            'protocols1_and_2' : ['first_accumulation_finished'],
            'protocol3_pretraining' : ['has_been_pretrained'],
            'protocol3_accumulation' : ['second_accumulation_finished'],
            'residual_identification' : ['has_been_assigned'],
            'post_processing' : ['has_crossings_solved', 'has_trajectories',
                                'has_trajectories_wo_gaps']}
    for process in processes:
        attributes = processes_to_attributes[process]
        attributes_values = []
        print(process)
        for attribute in attributes:
            print(attribute, getattr(old_video, attribute))
            attributes_values.append(getattr(old_video, attribute))
        if None in attributes_values:
            existentFile[process] = '-1'
        elif all(attributes_values):
            logger.debug(attribute)
            existentFile[process] = '1'
        else:
            existentFile[process] = '0'

    return existentFile

def getExistentFiles(video_object, processes):
    """get processes already computed in a previous session
    preprocessing: segmentation, fragmentation and creation of blobs and individual/global fragments
    knowledge_transfer: knowledge transferred from a model trained on a different video_object
    first_accumulation: first accumulation attempt
    pretraining: building the filters in a global-identity-agnostic way
    second_accumulation: accumulation by transferring knowledge from pre-training
    assignment: assignment of the idenitity to each individual fragment
    solving_duplications: solve eventual identity duplications
    crossings: assign identity to single animals during occlusions
    trajectories: compute the individual trajectories
    """
    existentFile = {name:'-1' for name in processes}
    old_video = None
    if os.path.isdir(video_object._previous_session_folder):
        logger.debug("loading old video object from get existent files")
        if os.path.isfile(os.path.join(video_object._previous_session_folder, 'video_object.npy')):
            print(video_object._previous_session_folder)
            old_video = np.load(os.path.join(video_object._previous_session_folder, 'video_object.npy'), allow_pickle=True).item()
            old_video.update_paths(os.path.join(video_object._previous_session_folder, 'video_object.npy'))
            logger.info("old video_object loaded")
        else:
            logger.info("The folder %s is empty. The tracking cannot be restored." %video_object._previous_session_folder)
            return existentFile, old_video

        old_video = check_and_change_video_path(video_object,old_video)
        existentFile = set_load_previous_dict(old_video, processes, existentFile)

    return existentFile, old_video

def interpolate_nans(t):
    """Interpolates nans linearly in a trajectory

    :param t: trajectory
    :returns: interpolated trajectory
    """
    shape_t = t.shape
    reshaped_t = t.reshape((shape_t[0], -1))
    for timeseries in range(reshaped_t.shape[-1]):
        y = reshaped_t[:, timeseries]
        nans, x = _nan_helper(y)
        y[nans] = np.interp(x(nans), x(~nans), y[~nans])

    # Ugly slow hack, as reshape seems not to return a view always
    back_t = reshaped_t.reshape(shape_t)
    t[...] = back_t

def _nan_helper(y):
    """Helper to handle indices and logical indices of NaNs.
    https://stackoverflow.com/questions/6518811/interpolate-nan-values-in-a-numpy-array
    Input:
        - y, 1d numpy array with possible NaNs
    Output:
        - nans, logical indices of NaNs
        - index, a function, with signature indices= index(logical_indices),
          to convert logical indices of NaNs to 'equivalent' indices
    Example:
        >>> # linear interpolation of NaNs
        >>> nans, x= nan_helper(y)
        >>> y[nans]= np.interp(x(nans), x(~nans), y[~nans])
    """

    return np.isnan(y), lambda z: z.nonzero()[0]



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import sys
import numpy as np
from confapp import conf

from idtrackerai.list_of_fragments import load_identification_images

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.globalfragment")

class GlobalFragment(object):
    """ A global fragment is a collection of instances of the class
    :class:`~fragment.Fragment`. Such fragments are collected from a part of the
    video in which all animals are visible.

    Attributes
    ----------

    index_beginning_of_fragment : int
        minimum frame number in which all the individual fragments (see
        :class:`~fragment.Fragment`) are all coexisting
    individual_fragments_identifiers : list
        list of the fragment identifiers associated to the individual fragments
        composing the global fragment (see :class:`~fragment.Fragment`)
    number_of_animals : int
        number of animals to be tracked
    _is_unique : bool
        True if each of the individual fragments have been assigned to a unique
        identity
    _is_certain : bool
        True if each of the individual fragments have scored a certaninty above
        the threshold :const:`conf.CERTAINTY_THRESHOLD`
    _ids_assigned : ndarray
        shape [1, number_of_animals] each componenents correspond to the
        identity assigned by the algorithm to each of the individual fragments
    _temporary_ids : ndarray
        shape [1, number_of_animals] temporary ids assigned during the
        fingerprinting protocol to each of the fragments composing the global
        fragment
    _repeated_ids : list
        list of identities repeated during the identification of the fragments
        in the individual fragment
    _missing_ids : list
        list of identities not assigned in the global fragment (since in a
        global fragment all animals are visible, all the identities should be
        assigned)
    predictions : list
        list of :attr:`~fragment.Fragment.predictions` for every individual
        fragment
    softmax_probs_median : list
        list of :attr:`~fragment.Fragment.softmax_probs_median` for every
        individual fragment

    """
    def __init__(self, blobs_in_video, fragments, index_beginning_of_fragment, number_of_animals):
        self.index_beginning_of_fragment = index_beginning_of_fragment
        self.individual_fragments_identifiers = [blob.fragment_identifier for blob in blobs_in_video[index_beginning_of_fragment]]
        self.get_list_of_attributes_from_individual_fragments(fragments)
        self.set_minimum_distance_travelled()
        self.set_candidate_for_accumulation()
        self.number_of_animals = number_of_animals
        self.reset(roll_back_to = 'fragmentation')
        self._is_unique = False
        self._is_certain = False

    @property
    def candidate_for_accumulation(self):
        return self._candidate_for_accumulation

    @property
    def used_for_training(self):
        return all([fragment.used_for_training for fragment in self.individual_fragments])

    @property
    def is_unique(self):
        self.check_uniqueness(scope = 'global')
        return self._is_unique

    @property
    def is_partially_unique(self):
        self.check_uniqueness(scope = 'partial')
        return self._is_partially_unique

    def reset(self, roll_back_to):
        """Resets attributes to the fragmentation step in the algorithm,
        allowing for example to start a new accumulation

        Parameters
        ----------
        roll_back_to : str
            "fragmentation"

        """
        if roll_back_to == 'fragmentation':
            self._ids_assigned = np.nan * np.ones(self.number_of_animals)
            self._temporary_ids = np.arange(self.number_of_animals)
            self._score = None
            self._is_unique = False
            self._uniqueness_score = None
            self._repeated_ids = []
            self._missing_ids = []
            self.predictions = []
            self.softmax_probs_median = []

    def get_individual_fragments_of_global_fragment(self, fragments):
        """Get the individual fragments in the global fragments by using their
        unique identifiers

        Parameters
        ----------
        fragments : list
            all the fragments extracted from the video
            (see :class:`~fragment.Fragment`)

        """
        self.individual_fragments = [fragment for fragment in fragments
                                        if fragment.identifier in self.individual_fragments_identifiers]

    def get_list_of_attributes_from_individual_fragments(self, fragments,
                list_of_attributes = ['distance_travelled', 'number_of_images']):
        """Given a set of attributes available in the instances of the class
        :class:`~fragment.Fragment` it copies them in the global fragment.
        For instance the attribute number_of_images belonging to the individual
        fragments in the global fragment will be set as
        `global_fragment.number_of_images_per_individual_fragment`, where each
        element of the list corresponds to the number_of_images of each
        individual fragment preserving the order with which the global fragment
        has been initialised

        Parameters
        ----------
        fragments : <Fragment object>
            See :class:`~fragment.Fragment`
        list_of_attributes : list
            List of attributes to be transferred from the individual fragments
            to the global fragment

        """
        [setattr(self, attribute + '_per_individual_fragment',[]) for attribute in list_of_attributes]
        for fragment in fragments:
            if fragment.identifier in self.individual_fragments_identifiers:
                assert fragment.is_an_individual
                setattr(fragment, '_is_in_a_global_fragment', True)
                for attribute in list_of_attributes:
                    getattr(self, attribute + '_per_individual_fragment').append(getattr(fragment, attribute))

    def set_minimum_distance_travelled(self):
        """Sets the minum distance travelled attribute
        """
        self.minimum_distance_travelled = min(self.distance_travelled_per_individual_fragment)

    def set_candidate_for_accumulation(self):
        """Sets the global fragment to be eligible for accumulation
        """
        self._candidate_for_accumulation = True
        if np.min(self.number_of_images_per_individual_fragment) < conf.MINIMUM_NUMBER_OF_FRAMES_TO_BE_A_CANDIDATE_FOR_ACCUMULATION:
            self._candidate_for_accumulation = False

    def get_total_number_of_images(self):
        return sum([fragment.number_of_images for fragment in self.individual_fragments])

    def acceptable_for_training(self, accumulation_strategy):
        """Returns True if the global fragment is acceptable for training.
        See :attr:`~fragment.Fragment.acceptable_for_training` for every
        individual fragment in the global fragment

        Parameters
        ----------
        accumulation_strategy : str
            can be either "global" or "partial"

        Returns
        -------
        bool
            True if the global fragment is accceptable for training

        """
        if accumulation_strategy == 'global':
            return all([fragment.acceptable_for_training for fragment in self.individual_fragments])
        else:
            return any([fragment.acceptable_for_training for fragment in self.individual_fragments])

    def check_uniqueness(self, scope):
        """Checks that the identities assigned to the individual fragments are
        unique

        Parameters
        ----------
        scope : str
            Either "global" or "partial"

        """
        all_identities = range(self.number_of_animals)
        if scope == 'global':
            if len(set(all_identities) - set([fragment.temporary_id for fragment in self.individual_fragments])) > 0:
                self._is_unique = False
            else:
                self._is_unique = True
        elif scope == 'partial':
            identities_acceptable_for_training = [fragment.temporary_id for fragment in self.individual_fragments
                                                    if fragment.acceptable_for_training]
            self.duplicated_identities = set([x for x in identities_acceptable_for_training if identities_acceptable_for_training.count(x) > 1])
            if len(self.duplicated_identities) > 0:
                self._is_partially_unique = False
            else:
                self._is_partially_unique = True

    def get_total_number_of_images(self):
        """Gets the total number of images in the global fragment
        """
        if not hasattr(self,'total_number_of_images'):
            self.total_number_of_images = sum([fragment.number_of_images for fragment in self.individual_fragments])
        return self.total_number_of_images

    def get_images_and_labels(self, identification_images_file_path, scope='pretraining'):
        """Arrange the images and identities in the global fragment as a
        labelled dataset in order to train the idCNN
        """
        images = []
        labels = []

        for temporary_id, fragment in enumerate(self.individual_fragments):
            images.extend(fragment.images)
            labels.extend([temporary_id] * fragment.number_of_images)
            if scope=='pretraining':
                fragment._temporary_id_for_pretraining = temporary_id

        return np.asarray(load_identification_images(identification_images_file_path, images)), labels

    # def compute_start_end_frame_indices_of_individual_fragments(self, blobs_in_video):
    #     """
    #
    #     Parameters
    #     ----------
    #     blobs_in_video : list
    #         list of the blob objects (see :class:`~blob.Blob`) segmented from
    #         the video
    #
    #     """
    #     self.starts_ends_individual_fragments = [blob.compute_fragment_start_end()
    #         for blob in blobs_in_video[self.index_beginning_of_fragment]]

    def update_individual_fragments_attribute(self, attribute, value):
        """Update `attribute` in every individual fragment in the global
        fragment by setting it at `value`

        Parameters
        ----------
        attribute : str
            attribute to be updated
        value : list, int, float
            value of `attribute`

        """
        [setattr(fragment, attribute, value) for fragment in self.individual_fragments]



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import sys

import numpy as np

from confapp import conf

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.epoch_runner_crossings")

class EpochRunner(object):
    def __init__(self, data_set,
                starting_epoch = 0,
                print_flag = False):
        """Runs training a model given the network and the data set
        """
        # Counters
        self._epochs_completed = 0
        self.starting_epoch= starting_epoch
        # Number of epochs
        self.print_flag = print_flag
        # Data set
        self.data_set = data_set
        self._num_images = len(data_set.images)

    @property
    def epochs_completed(self):
        return self._epochs_completed

    def next_batch(self, batch_size):
        """Return the next `batch_size` examples from this data set."""
        start = self._index_in_epoch
        self._index_in_epoch += batch_size
        end = self._index_in_epoch
        return (self.data_set.images[start:end], self.data_set.labels[start:end])

    def run_epoch(self, name, store_loss_and_accuracy, batch_operation):
        loss_epoch = []
        accuracy_epoch = []
        individual_accuracy_epoch = []
        self._index_in_epoch = 0
        while self._index_in_epoch < self._num_images:
            loss_acc_batch, feed_dict = batch_operation(self.next_batch(conf.BATCH_SIZE_DCD))
            loss_epoch.append(loss_acc_batch[0])
            accuracy_epoch.append(loss_acc_batch[1])
            individual_accuracy_epoch.append(loss_acc_batch[2])
        loss_epoch = np.mean(np.vstack(loss_epoch))
        accuracy_epoch = np.mean(np.vstack(accuracy_epoch))
        individual_accuracy_epoch = np.nanmean(np.vstack(individual_accuracy_epoch),axis=0)
        if self.print_flag:
            logger.debug(name)
            logger.debug('epoch: %i' %(self.starting_epoch + self._epochs_completed))
            logger.debug('loss: %s' %str(loss_epoch))
            logger.debug('accuracy: %s' %str(accuracy_epoch))
            logger.debug('individual accuray: %s' %str(individual_accuracy_epoch))
        store_loss_and_accuracy.append_data(loss_epoch, accuracy_epoch, individual_accuracy_epoch)
        return feed_dict
